"""Contains code for an Open Pixel Controller hardware for RGB LEDs.

The python code to build the OPC message packet came from here:
https://github.com/zestyping/openpixelcontrol/blob/master/python_clients/opc.py
"""
from typing import Optional

import logging

from mpf.core.platform import LightsPlatform
from mpf.platforms.interfaces.light_platform_interface import LightPlatformInterface

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


class OpenpixelHardwarePlatform(LightsPlatform):

    """Base class for the open pixel hardware platform.

    Args:
    ----
        machine: The main ``MachineController`` object.

    """

    __slots__ = ["opc_client"]

    def __init__(self, machine: "MachineController") -> None:
        """Instantiate openpixel hardware platform."""
        super().__init__(machine)

        self.log = logging.getLogger("OpenPixel")
        self.log.debug("Configuring Open Pixel hardware interface.")
        self.opc_client = None      # type: Optional[OpenPixelClient]
        self.features['tickless'] = True

    def __repr__(self):
        """Return str representation."""
        return '<Platform.OpenPixel>'

    async def initialize(self):
        """Initialise openpixel platform."""
        self.machine.config_validator.validate_config("open_pixel_control", self.machine.config['open_pixel_control'])
        if self.machine.config['open_pixel_control']['debug']:
            self.debug = True

        await self._setup_opc_client()

    def stop(self):
        """Stop platform."""
        # disconnect sender
        if self.opc_client:
            self.opc_client.blank_all()
            if self.opc_client.socket_sender:
                self.opc_client.socket_sender.close()
                self.opc_client.socket_sender = None
            self.opc_client = None

    def parse_light_number_to_channels(self, number: str, subtype: str):
        """Parse number to three channels."""
        del subtype
        if isinstance(number, str) and '-' in number:
            opc_channel, led_number = number.split('-')
        else:
            led_number = number
            opc_channel = "0"

        channel_number = int(led_number) * 3

        return [
            {
                "number": opc_channel + "-" + str(channel_number)
            },
            {
                "number": opc_channel + "-" + str(channel_number + 1)
            },
            {
                "number": opc_channel + "-" + str(channel_number + 2)
            }
        ]

    def configure_light(self, number, subtype, config, platform_settings) -> LightPlatformInterface:
        """Configure an LED."""
        del config
        return OpenPixelLED(number, self.opc_client, self.debug)

    async def _setup_opc_client(self):
        self.opc_client = OpenPixelClient(self.machine, self.machine.config['open_pixel_control'])
        await self.opc_client.connect()


class OpenPixelLED(LightPlatformInterface):

    """One LED on the openpixel platform."""

    __slots__ = ["log", "opc_client", "debug", "opc_channel", "channel_number"]

    def __init__(self, number, opc_client, debug):
        """Initialise Openpixel LED obeject."""
        super().__init__(number)
        self.log = logging.getLogger('OpenPixelLED')
        channel, channel_number = number.split("-")

        self.opc_client = opc_client
        self.debug = debug
        self.opc_channel = int(channel)
        self.channel_number = int(channel_number)
        self.opc_client.add_pixel(self.opc_channel, self.channel_number)

    def set_fade(self, start_brightness, start_time, target_brightness, target_time):
        """Set brightness using callback."""
        self.opc_client.set_pixel_color(self.opc_channel, self.channel_number, start_brightness, start_time,
                                        target_brightness, target_time)

    def get_board_name(self):
        """Return name for service mode."""
        return "OPC Channel {}".format(self.channel_number)

    def __repr__(self):
        """Return string representation."""
        return "<OpenPixelLED channel={} number={}>".format(self.opc_channel, self.channel_number)

    def is_successor_of(self, other):
        """Return true if the other light has the previous index and is on the same channel."""
        return self.opc_channel == other.opc_channel and self.channel_number == other.channel_number + 1

    def get_successor_number(self):
        """Return next index on node."""
        return "{}-{}".format(self.opc_channel, self.channel_number + 1)

    def __lt__(self, other):
        """Order lights by their order on the hardware."""
        return (self.opc_channel, self.channel_number) < (other.opc_channel, other.channel_number)


class OpenPixelClient:

    """Base class of an OPC client which connects to a FadeCandy server.

    Args:
    ----
        machine: The main ``MachineController`` instance.
        config: Config to use
    """

    __slots__ = ["machine", "log", "update_every_tick", "socket_sender", "max_fade_ms", "channels", "dirty_leds",
                 "msg", "openpixel_config"]

    def __init__(self, machine, config):
        """Initialise openpixel client."""
        self.log = logging.getLogger('OpenPixelClient')

        self.machine = machine
        self.update_every_tick = False
        self.socket_sender = None
        self.max_fade_ms = None
        self.channels = []
        self.dirty_leds = []
        self.msg = []
        self.openpixel_config = config

    async def connect(self):
        """Connect to the hardware."""
        connector = self.machine.clock.open_connection(self.openpixel_config['host'], self.openpixel_config['port'])
        try:
            _, self.socket_sender = await connector
        except OSError:
            raise AssertionError("Cannot connect to {} at {}:{}".format(self.log.name, self.openpixel_config['host'],
                                                                        self.openpixel_config['port']))

        self.max_fade_ms = int(1 / self.machine.config['mpf']['default_light_hw_update_hz'] * 1000)

        self.machine.events.add_handler("init_phase_3", self._start_loop)

    def _start_loop(self, **kwargs):
        del kwargs
        # blank all channels
        self.blank_all()

        # Update at a regular interval
        self.machine.clock.schedule_interval(self.tick, 1 / self.machine.config['mpf']['default_light_hw_update_hz'])

    def add_pixel(self, channel, led):
        """Add a pixel to the list that will be sent to the OPC server.

        Args:
        ----
            channel: Integer of the OPC channel this pixel is on.
            led: Integer of the pixel number (i.e. its position in the list) for
                this pixel.

        This is needed since MPF will process LED device entries in random
        order, so if (for example) we get a call to setup LED #20, then we need
        we make sure we have 19 items on the list before it.

        """
        if len(self.channels) < channel + 1:
            channels_to_add = channel + 1 - len(self.channels)

            self.channels += [list() for _ in range(channels_to_add)]
            self.dirty_leds += [dict() for _ in range(channels_to_add)]
            self.msg += [None for _ in range(channels_to_add)]

        new_total = led + 1
        if new_total % 3 != 0:
            new_total += 3 - (new_total % 3)
        leds_to_add = new_total - len(self.channels[channel])
        if leds_to_add > 0:
            self.channels[channel] += [0 for _ in range(leds_to_add)]

    # pylint: disable-msg=too-many-arguments
    def set_pixel_color(self, channel, pixel, start_brightness, start_time, target_brightness, target_time):
        """Set an individual pixel color.

        Args:
        ----
            channel: Int of the OPC channel for this pixel.
            pixel: Int of the number for this pixel on that channel.
            start_brightness: Brightness at start of fade.
            start_time: Timestamp when the fade started.
            target_brightness: Brightness at end of fade.
            target_time: Timestamp when the fade should finish.
        """
        self.dirty_leds[channel][pixel] = (start_brightness, start_time, target_brightness, target_time)

    def tick(self):
        """Update pixels.

        Called periodically.
        """
        for channel_index in range(len(self.channels)):
            if not self.update_every_tick and not self.dirty_leds[channel_index]:
                continue
            self._handle_dirty_leds(channel_index)
            self._update_pixels(channel_index)

    def _handle_dirty_leds(self, channel):
        if not self.dirty_leds[channel]:
            return

        # invalidate cached message
        self.msg[channel] = None

        current_time = self.machine.clock.get_time()
        max_fade_ms = self.max_fade_ms
        for pixel, (start_brightness, start_time, target_brightness, target_time) \
                in dict(self.dirty_leds[channel]).items():
            fade_ms = int((target_time - current_time) * 1000.0)
            if fade_ms > max_fade_ms > 0:
                ratio = ((current_time + (max_fade_ms / 1000.0) - start_time) /
                         (target_time - start_time))
                brightness = start_brightness + (target_brightness - start_brightness) * ratio
            else:
                # fade is done
                brightness = target_brightness
                del self.dirty_leds[channel][pixel]
            value = min(255, max(0, int(brightness * 255)))
            self.channels[channel][pixel] = value

    def _update_pixels(self, channel):
        """Send the list of pixel colors to the OPC server.

        Args:
        ----
            channel: Which OPC channel the pixel data will be written to.

        Returns True on success. False if it was unable to connect to the OPC
        server.

        Note that you must send color data for all the pixels in a channel (or
        all the pixels up until the point you want. e.g. if you have 30 LEDs on
        the channel and you just want to update LED #10, then you need to send
        pixel data for the first 10 pixels.)
        """
        # if we got a cached message just send it
        if not self.msg[channel]:
            self.msg[channel] = bytes(self._build_message(channel))

        self.send(self.msg[channel])

    def _build_message(self, channel):
        """Build the OPC message."""
        msg = bytearray()
        pixels = self.channels[channel]
        len_hi_byte = int(len(pixels) / 256)
        len_lo_byte = (len(pixels)) % 256
        header = bytes([channel, 0, len_hi_byte, len_lo_byte])
        msg.extend(header)
        for i in range(int(len(pixels) / 3)):
            # send GRB because that is the default color order for WS2812

            msg.append(pixels[i * 3 + 1])
            msg.append(pixels[i * 3])
            msg.append(pixels[i * 3 + 2])
        return msg

    def blank_all(self):
        """Blank all channels."""
        for channel_index in range(len(self.channels)):
            self.channels[channel_index] = [0] * len(self.channels[channel_index])
            self.send(bytes(self._build_message(channel_index)))

    def send(self, message):
        """Send a message to the socket.

        Args:
        ----
            message: Message to send
        """
        if not self.socket_sender:
            return
        self.socket_sender.write(message)



import asyncio

from mpf.tests.MpfTestCase import MpfTestCase, patch


class TestMMA8451(MpfTestCase):

    def get_platform(self):
        return False

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/mma8451/'

    async def i2c_read8(self, register):
        return self.i2c_layout[register]

    async def i2c_read_block(self, register, count):
        assert count == 6
        assert register == 0x01
        return self.read_value

    def i2c_write8(self, register, value):
        """Write to I2C."""
        key = (register, value)
        if key not in self.i2c_expect:
            raise AssertionError("Did not expect write to register {:02X} with value {:02X} ({})".
                                 format(register, value, key))
        del self.i2c_expect[key]

    def setUp(self):
        self.read_value = bytearray([0, 0, 0, 0, 0, 0])
        self.i2c_layout = {0x0D: 0x1A,      # ID of the device
                           0x2B: 00,        # reset success
                           }
        self.i2c_expect = {(0x2B, 0x40): True,  # reset
                           (0x2B, 0x02): True,  # resolution
                           (0x2D, 0x01): True,  # ready true
                           (0x2E, 0x01): True,  # ready true
                           (0x11, 0x40): True,  # orientation mode on
                           (0x2A, 0x2D): True,  # low noise and activate
                           }
        with patch("mpf.platforms.virtual.VirtualI2cDevice.i2c_read8", new=self.i2c_read8):
            with patch("mpf.platforms.virtual.VirtualI2cDevice.i2c_read_block", new=self.i2c_read_block):
                with patch("mpf.platforms.virtual.VirtualI2cDevice.i2c_write8", new=self.i2c_write8):
                    super().setUp()
                    self.advance_time_and_run(1)

        self.assertFalse(self.i2c_expect)

    def test_init_and_poll(self):
        with patch("mpf.platforms.virtual.VirtualI2cDevice.i2c_read8", new=self.i2c_read8):
            with patch("mpf.platforms.virtual.VirtualI2cDevice.i2c_read_block", new=self.i2c_read_block):
                with patch("mpf.platforms.virtual.VirtualI2cDevice.i2c_write8", new=self.i2c_write8):
                    self.assertEqual((0, 0, 0), self.machine.accelerometers["test_accelerometer"].value)

                    self.read_value = bytearray([0, 0, 0, 0, 60, 10])
                    self.advance_time_and_run(.1)

                    self.assertEqual((0, 0, 9.199), self.machine.accelerometers["test_accelerometer"].value)



"""Coil config player."""
from copy import deepcopy
from typing import List

from mpf.config_players.device_config_player import DeviceConfigPlayer
from mpf.devices.driver import Driver


class CoilPlayer(DeviceConfigPlayer):

    """Triggers coils based on config."""

    config_file_section = 'coil_player'
    show_section = 'coils'
    machine_collection_name = 'coils'

    __slots__ = []  # type: List[str]

    def play(self, settings, context: str, calling_context: str, priority: int = 0, **kwargs):
        """Enable, Pulse or disable coils."""
        del kwargs
        del calling_context
        instance_dict = self._get_instance_dict(context)

        for coil, s in settings.items():
            s = deepcopy(s)
            if not isinstance(coil, Driver):
                self.raise_config_error("Invalid coil name {}".format(coil), 2, context=context)
            action = s.pop('action')

            # delete coil from dict
            try:
                del instance_dict[coil.name]
            except KeyError:
                pass

            if action in ("disable", "off"):
                coil.disable()
            elif action in ("on", "enable"):
                instance_dict[coil.name] = coil
                coil.enable(pulse_ms=s["pulse_ms"], pulse_power=s["pulse_power"], hold_power=s["hold_power"])
            elif action == "pulse":
                coil.pulse(pulse_ms=s['pulse_ms'], pulse_power=s['pulse_power'], max_wait_ms=s['max_wait_ms'])
            else:
                self.raise_config_error("Invalid action {}".format(action), 1, context=context)

    def clear_context(self, context):
        """Disable enabled coils."""
        instance_dict = self._get_instance_dict(context)
        for coil in instance_dict.values():
            coil.disable()

        self._reset_instance_dict(context)

    def get_express_config(self, value: str):
        """Parse short config version."""
        try:
            int_value = int(value)
            return dict(action='pulse', pulse_ms=int_value)
        except (TypeError, ValueError):
            pass

        action = 'pulse'

        if value in ('disable', 'off'):
            action = 'disable'

        elif value in ('enable', 'on'):
            action = 'enable'

        return dict(action=action)



from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase


class TestBallDevicePlayfieldLock(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'test_playfield_lock.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def get_platform(self):
        return 'smart_virtual'

    def test_entry_during_eject(self):
        self.start_game()
        self.mock_event("balldevice_test_device_ball_eject_failed")
        self.mock_event("balldevice_test_device_ball_eject_success")

        # lock a ball
        self.hit_switch_and_run("s_ball_switch1", 10)
        self.assertEqual(1, self.machine.ball_devices["test_device"].available_balls)

        # new ball rolls in
        self.post_event("entrance_event")

        # release ball
        self.post_event("release_test")
        self.advance_time_and_run(.1)
        self.assertEqual(0, self.machine.ball_devices["test_device"].available_balls)

        # capture a new ball concurrently
        self.hit_switch_and_run("s_ball_switch2", 20)
        self.assertEventNotCalled("balldevice_test_device_ball_eject_failed")
        self.assertEventCalled("balldevice_test_device_ball_eject_success")
        self.mock_event("balldevice_test_device_ball_eject_failed")
        self.mock_event("balldevice_test_device_ball_eject_success")

        # locked another ball
        self.assertEqual(1, self.machine.ball_devices["test_device"].available_balls)

        # release another ball
        self.post_event("release_test")
        self.advance_time_and_run(.1)
        self.assertEqual(0, self.machine.ball_devices["test_device"].available_balls)

        # no entrance but a new ball
        self.hit_switch_and_run("s_ball_switch2", 20)
        self.assertEventCalled("balldevice_test_device_ball_eject_failed")
        self.assertEventNotCalled("balldevice_test_device_ball_eject_success")




"""The Color module provides utilities for working with RGB colors.

It is based on the colorutils open-source library:
https://github.com/edaniszewski/colorutils
Copyright (c) 2015 Erick Daniszewski
The MIT License (MIT)
"""
import random

from typing import List, Union, Tuple, Optional

from mpf.core.utility_functions import Util

CHANNEL_MIN_VAL = 0
CHANNEL_MAX_VAL = 255
RGB_MIN = (0, 0, 0)
RGB_MAX = (255, 255, 255)

# Standard web color names and values
NAMED_RGB_COLORS = dict(
    off=(0, 0, 0),
    aliceblue=(240, 248, 255),
    antiquewhite=(250, 235, 215),
    aqua=(0, 255, 255),
    aquamarine=(127, 255, 212),
    azure=(240, 255, 255),
    beige=(245, 245, 220),
    bisque=(255, 228, 196),
    black=(0, 0, 0),
    blanchedalmond=(255, 235, 205),
    blue=(0, 0, 255),
    blueviolet=(138, 43, 226),
    brown=(165, 42, 42),
    burlywood=(222, 184, 135),
    cadetblue=(95, 158, 160),
    chartreuse=(127, 255, 0),
    chocolate=(210, 105, 30),
    coral=(255, 127, 80),
    cornflowerblue=(100, 149, 237),
    cornsilk=(255, 248, 220),
    crimson=(220, 20, 60),
    cyan=(0, 255, 255),
    darkblue=(0, 0, 139),
    darkcyan=(0, 139, 139),
    darkgoldenrod=(184, 134, 11),
    darkgray=(169, 169, 169),
    darkgreen=(0, 100, 0),
    darkkhaki=(189, 183, 107),
    darkmagenta=(139, 0, 139),
    darkolivegreen=(85, 107, 47),
    darkorange=(255, 140, 0),
    darkorchid=(153, 50, 204),
    darkred=(139, 0, 0),
    darksalmon=(233, 150, 122),
    darkseagreen=(143, 188, 143),
    darkslateblue=(72, 61, 139),
    darkslategray=(47, 79, 79),
    darkturquoise=(0, 206, 209),
    darkviolet=(148, 0, 211),
    deeppink=(255, 20, 147),
    deepskyblue=(0, 191, 255),
    dimgray=(105, 105, 105),
    dodgerblue=(30, 144, 255),
    firebrick=(178, 34, 34),
    floralwhite=(255, 250, 240),
    forestgreen=(34, 139, 34),
    fuchsia=(255, 0, 255),
    gainsboro=(220, 220, 220),
    ghostwhite=(248, 248, 255),
    gold=(255, 215, 0),
    goldenrod=(218, 165, 32),
    gray=(128, 128, 128),
    green=(0, 128, 0),
    greenyellow=(173, 255, 47),
    honeydew=(240, 255, 240),
    hotpink=(255, 105, 180),
    indianred=(205, 92, 92),
    indigo=(75, 0, 130),
    ivory=(255, 255, 240),
    khaki=(240, 230, 140),
    lavender=(230, 230, 250),
    lavenderblush=(255, 240, 245),
    lawngreen=(124, 252, 0),
    lemonchiffon=(255, 250, 205),
    lightblue=(173, 216, 230),
    lightcoral=(240, 128, 128),
    lightcyan=(224, 255, 255),
    lightgoldenrodyellow=(250, 250, 210),
    lightgreen=(144, 238, 144),
    lightgrey=(211, 211, 211),
    lightpink=(255, 182, 193),
    lightsalmon=(255, 160, 122),
    lightseagreen=(32, 178, 170),
    lightskyblue=(135, 206, 250),
    lightslategray=(119, 136, 153),
    lightsteelblue=(176, 196, 222),
    lightyellow=(255, 255, 224),
    lime=(0, 255, 0),
    limegreen=(50, 205, 50),
    linen=(250, 240, 230),
    magenta=(255, 0, 255),
    maroon=(128, 0, 0),
    mediumaquamarine=(102, 205, 170),
    mediumblue=(0, 0, 205),
    mediumorchid=(186, 85, 211),
    mediumpurple=(147, 112, 219),
    mediumseagreen=(60, 179, 113),
    mediumslateblue=(123, 104, 238),
    mediumspringgreen=(0, 250, 154),
    mediumturquoise=(72, 209, 204),
    mediumvioletred=(199, 21, 133),
    midnightblue=(25, 25, 112),
    mintcream=(245, 255, 250),
    mistyrose=(255, 228, 225),
    moccasin=(255, 228, 181),
    navajowhite=(255, 222, 173),
    navy=(0, 0, 128),
    oldlace=(253, 245, 230),
    olive=(128, 128, 0),
    olivedrab=(107, 142, 35),
    orange=(255, 165, 0),
    orangered=(255, 69, 0),
    orchid=(218, 112, 214),
    palegoldenrod=(238, 232, 170),
    palegreen=(152, 251, 152),
    paleturquoise=(175, 238, 238),
    palevioletred=(219, 112, 147),
    papayawhip=(255, 239, 213),
    peachpuff=(255, 218, 185),
    peru=(205, 133, 63),
    pink=(255, 192, 203),
    plum=(221, 160, 221),
    powderblue=(176, 224, 230),
    purple=(128, 0, 128),
    rebeccapurple=(102, 51, 153),
    red=(255, 0, 0),
    rosybrown=(188, 143, 143),
    royalblue=(65, 105, 225),
    saddlebrown=(139, 69, 19),
    salmon=(250, 128, 114),
    sandybrown=(244, 164, 96),
    seagreen=(46, 139, 87),
    seashell=(255, 245, 238),
    sienna=(160, 82, 45),
    silver=(192, 192, 192),
    skyblue=(135, 206, 235),
    slateblue=(106, 90, 205),
    slategray=(112, 128, 144),
    snow=(255, 250, 250),
    springgreen=(0, 255, 127),
    steelblue=(70, 130, 180),
    tan=(210, 180, 140),
    teal=(0, 128, 128),
    thistle=(216, 191, 216),
    tomato=(255, 99, 71),
    turquoise=(64, 224, 208),
    violet=(238, 130, 238),
    wheat=(245, 222, 179),
    white=(255, 255, 255),
    whitesmoke=(245, 245, 245),
    yellow=(255, 255, 0),
    yellowgreen=(154, 205, 50),
)


class RGBColor:

    """One RGB Color."""

    __slots__ = ["_color"]

    def __init__(self, color: Union["RGBColor", str, List[int], Tuple[int, int, int]] = None) -> None:
        """Initialise color."""
        if isinstance(color, RGBColor):
            self._color = color.rgb
        elif isinstance(color, str):
            self._color = RGBColor.string_to_rgb(color)
        elif color:
            self._color = (color[0], color[1], color[2])
        else:
            self._color = RGB_MIN

    def __eq__(self, other):
        """Return true if equal."""
        return RGBColor(other).rgb == self.rgb

    def __ne__(self, other):
        """Return true if not equal."""
        return not self.__eq__(other)

    def __add__(self, other):
        """Return sum of two RGB colors."""
        if isinstance(other, RGBColor):
            r1, g1, b1 = self.rgb
            r2, g2, b2 = other.rgb
        elif isinstance(other, tuple) and len(other) == 3:
            r1, g1, b1 = self.rgb
            r2, g2, b2 = other
        else:
            raise TypeError(
                "Unsupported operand type(s) for +: '{0}' and '{1}'".format(
                    type(self), type(other)))

        return RGBColor((min(r1 + r2, CHANNEL_MAX_VAL),
                         min(g1 + g2, CHANNEL_MAX_VAL),
                         min(b1 + b2, CHANNEL_MAX_VAL)))

    def __sub__(self, other):
        """Return difference of two RGB colors."""
        if isinstance(other, RGBColor):
            r1, g1, b1 = self.rgb
            r2, g2, b2 = other.rgb
        elif isinstance(other, tuple) and len(other) == 3:
            r1, g1, b1 = self.rgb
            r2, g2, b2 = other
        else:
            raise TypeError(
                "Unsupported operand type(s) for -: '{0}' and '{1}'".format(
                    type(self), type(other)))

        return RGBColor((max(r1 - r2, CHANNEL_MIN_VAL),
                         max(g1 - g2, CHANNEL_MIN_VAL),
                         max(b1 - b2, CHANNEL_MIN_VAL)))

    def __mul__(self, other):
        """Multiple color by scalar."""
        if not isinstance(other, (float, int)):
            raise TypeError(
                "Unsupported operand type(s) for *: '{0}' and '{1}'".format(
                    type(self), type(other)))

        if other < 0:
            raise TypeError("Operand needs to be positive")

        r1, g1, b1 = self.rgb
        return RGBColor((min(int(r1 * other), CHANNEL_MAX_VAL),
                         min(int(g1 * other), CHANNEL_MAX_VAL),
                         min(int(b1 * other), CHANNEL_MAX_VAL)))

    def __iter__(self):
        """Return iterator."""
        return iter(self._color)

    def __str__(self):
        """Return string representation."""
        return "{}".format(self._color)

    def __repr__(self):
        """Return general representation."""
        return "<RGBColor {} ({})>".format(self._color, self.name)

    @property
    def red(self):
        """Return the red component of the RGB color representation."""
        return self._color[0]

    @property
    def green(self) -> int:
        """Return the green component of the RGB color representation."""
        return self._color[1]

    @property
    def blue(self) -> int:
        """Return the blue component of the RGB color representation."""
        return self._color[2]

    @property
    def rgb(self) -> Tuple[int, int, int]:
        """Return an RGB representation of the color."""
        return self._color

    @property
    def hex(self) -> str:
        """Return a 6-char HEX representation of the color."""
        return RGBColor.rgb_to_hex(self.rgb)

    @property
    def name(self) -> Optional[str]:
        """Return the color name or None.

        Returns a string containing a standard color name or None
        if the current RGB color does not have a standard name.
        """
        # pylint: disable-msg=consider-using-dict-comprehension
        return dict(
            [(_v, _k) for _k, _v in list(NAMED_RGB_COLORS.items())]).get(
            self._color)

    @staticmethod
    def rgb_to_hex(rgb: Tuple[int, int, int]) -> str:
        """Convert an RGB color representation to a HEX color representation.

        (r, g, b) :: r -> [0, 255]
                     g -> [0, 255]
                     b -> [0, 255]
        :param rgb: A tuple of three numeric values corresponding to the red, green, and blue value.
        :return: HEX representation of the input RGB value.
        :rtype: str
        """
        r, g, b = rgb
        return "{0}{1}{2}".format(hex(int(r))[2:].zfill(2),
                                  hex(int(g))[2:].zfill(2),
                                  hex(int(b))[2:].zfill(2))

    @staticmethod
    def hex_to_rgb(hex_string: str, default=None) -> Tuple[int, int, int]:
        """Convert a HEX color representation to an RGB color representation.

        Args:
        ----
            hex_string: The 3- or 6-char hexadecimal string representing the color
                value.
            default: The default value to return if _hex is invalid.

        Returns: RGB representation of the input HEX value as a 3-item tuple
            with each item being an integer 0-255.

        """
        if not Util.is_hex_string(hex_string):
            return default

        hex_string = str(hex_string).strip('#')

        n = len(hex_string) // 3
        r = int(hex_string[:n], 16)
        g = int(hex_string[n:2 * n], 16)
        b = int(hex_string[2 * n:3 * n], 16)
        return r, g, b

    @staticmethod
    def blend(start_color, end_color, fraction):
        """Blend two colors.

        Args:
        ----
            start_color: The start color
            end_color:  The end color
            fraction: The fraction between 0 and 1 that is used to set the
                blend point between the two colors.

        Returns: An RGBColor object that is a blend between the start and end
            colors

        """
        if isinstance(start_color, RGBColor):
            start_color = start_color.rgb
        else:
            start_color = RGBColor(start_color).rgb

        if isinstance(end_color, RGBColor):
            end_color = end_color.rgb
        else:
            end_color = RGBColor(start_color).rgb

        output_color = tuple(start_color[i] + int(
            (end_color[i] - start_color[i]) * fraction) for i in range(3))
        return RGBColor(output_color)

    @staticmethod
    def random_rgb() -> Tuple[int, int, int]:
        """Generate a uniformly random RGB value.

        :return: A tuple of three integers with values between 0 and 255 inclusive
        """
        return random.randint(0, 255), random.randint(0, 255), random.randint(
            0, 255)

    @staticmethod
    def name_to_rgb(name: str, default=RGB_MIN) -> Tuple[int, int, int]:
        """Convert a standard color name to an RGB value (tuple).

        If the name is not found, the default value is returned.
        :param name: A standard color name.
        :param default: The default value to return if the color name is not found.
        :return: RGB representation of the named color.
        :rtype: tuple
        """
        return NAMED_RGB_COLORS.get(name.lower(), default)

    @staticmethod
    def string_to_rgb(value: str, default=RGB_MIN) -> Tuple[int, int, int]:
        """Convert a string which could be either a standard color name or a hex value to an RGB value (tuple).

        If the name is not found and the supplied value is not a
        valid hex string it raises an error.
        :param value: A standard color name or hex value.
        :param default: The default value to return if the color name is not found and
        the supplied value is not a valid hex color string.
        :return: RGB representation of the named color.
        :rtype: tuple
        """
        if not value:
            return default

        brightness = None
        if '%' in value:
            value, brightness = value.split("%")

        rgb = NAMED_RGB_COLORS.get(value)
        if rgb is None:
            # we do not want to call lower every time since this code path is very hot
            # instead we just add the upper case string to the color hash map so next time we will hit the fast path
            rgb = NAMED_RGB_COLORS.get(value.lower())
            if rgb:
                NAMED_RGB_COLORS[value.lower()] = rgb
        if rgb is None:
            rgb = RGBColor.hex_to_rgb(value)
            if rgb is None:
                raise ColorException("Invalid RGB string: {}".format(value))

        # apply brightness
        if brightness:
            factor = float(int(brightness) / 100)
            rgb = (min(int(rgb[0] * factor), CHANNEL_MAX_VAL),
                   min(int(rgb[1] * factor), CHANNEL_MAX_VAL),
                   min(int(rgb[2] * factor), CHANNEL_MAX_VAL))

        return rgb

    @staticmethod
    def add_color(name: str, color: Union["RGBColor", str, List[int], Tuple[int, int, int]]):
        """Add (or updates if it already exists) a color.

        Note that this is not
        permanent, the list is reset when MPF restarts (though you can define your
        own custom colors in your config file's colors: section). You *can* use
        this function to dynamically change the values of colors in shows (they
        take place the next time an LED switches to that color).

        Args:
        ----
            name: String name of the color you want to add/update
            color: The color you want to set. You can pass the same types as
                the RGBColor class constructor, including a tuple or list of
                RGB ints (0-255 each), a hex string, an RGBColor instance, or
                a dictionart of red, green, blue key/value pairs.

        """
        NAMED_RGB_COLORS[str(name.lower())] = RGBColor(color).rgb


class ColorException(AssertionError):

    """General exception thrown for color utilities non-exit exceptions."""


class RGBColorCorrectionProfile:

    """Encapsulates a named RGB color correction profile and its associated lookup tables."""

    def __init__(self, name: str) -> None:
        """Create a linear correction profile that does not alter color values by default.

        Args:
        ----
            name: The color correction profile name

        Returns: None
        """
        self._name = name

        # Default lookup table values (linear)
        self._lookup_table = []             # type: List[List[int]]

        for dummy_channel in range(3):
            self._lookup_table.append(list(range(256)))

    def generate_from_parameters(self, gamma=2.5, whitepoint=(1.0, 1.0, 1.0),
                                 linear_slope=1.0, linear_cutoff=0.0):
        """Generate an RGB color correction profile lookup table based on the parameters supplied.

        Args:
        ----
            gamma: Exponent for the nonlinear portion of the brightness curve.
            whitepoint: Tuple of (red, green, blue) values to multiply by
                colors prior to gamma correction.
            linear_slope: Slope (output / input) of the linear section of the
                brightness curve.
            linear_cutoff: Y (output) coordinate of intersection between linear
                and nonlinear curves.

        Returns: None
        """
        # Lookup table generation algorithm from the Fadecandy open source
        # server code:
        # https://github.com/scanlime/fadecandy
        # Copyright (c) 2013 Micah Elizabeth Scott
        # The MIT License (MIT)
        scale = 1.0 - linear_cutoff

        for channel in range(3):
            for index in range(256):
                # Scale linear table values by the whitepoint
                value = index / 255.0 * whitepoint[channel]

                if value * linear_slope <= linear_cutoff:
                    value = int(linear_slope * value * 255)
                else:
                    non_linear_input = value - (linear_slope * linear_cutoff)
                    value = int(linear_cutoff + pow(non_linear_input / scale,
                                                    gamma) * scale * 255)

                # Clamp the lookup table value between 0 and 255
                self._lookup_table[channel][index] = max(0, min(value, 255))

    def assign_channel_lookup_table_values(self, channel: int, table_values: List[int]):
        """Assign the specified lookup table values to the profile channel.

        Args:
        ----
            channel: The channel number (0..2)
            table_values: A list of 256 integer values between 0 and 255

        """
        if channel not in range(3):
            raise ValueError('Invalid channel number in color correction profile')

        if not isinstance(table_values, list) or len(table_values) != 256:
            raise TypeError('Invalid lookup table values type for color correction profile - '
                            'must be a list of 256 integer values')

        for index in range(256):
            value = table_values[index]
            if not isinstance(value, int) or value < 0 or value > 255:
                raise ValueError('Invalid value in color correction profile channel lookup table - '
                                 'must be integers between 0 and 255')

            self._lookup_table[channel][index] = value

    @property
    def name(self) -> str:
        """Return the color correction profile name.

        Returns str name of color.
        """
        return self._name

    def apply(self, color) -> RGBColor:
        """Apply the current color correction profile to the specified RGBColor object.

        Args:
        ----
            color: The RGBColor object which to apply the color correction profile.

        Returns: RGBColor
        """
        assert 0 <= color.red <= 255
        assert 0 <= color.green <= 255
        assert 0 <= color.blue <= 255
        assert len(self._lookup_table[0]) == 256
        assert len(self._lookup_table[1]) == 256
        assert len(self._lookup_table[2]) == 256
        return RGBColor((self._lookup_table[0][color.red],
                         self._lookup_table[1][color.green],
                         self._lookup_table[2][color.blue]))

    @staticmethod
    def default() -> "RGBColorCorrectionProfile":
        """Create a default profile (gamma-corrected).

        The values for this table come from a web article:
        https://learn.adafruit.com/led-tricks-gamma-correction/the-quick-fix
        """
        default_profile = RGBColorCorrectionProfile('default')

        # Create standard gamma correction lookup table values
        table = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1,
                 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2,
                 2, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 5, 5, 5,
                 5, 6, 6, 6, 6, 7, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10,
                 10, 10, 11, 11, 11, 12, 12, 13, 13, 13, 14, 14, 15, 15, 16, 16,
                 17, 17, 18, 18, 19, 19, 20, 20, 21, 21, 22, 22, 23, 24, 24, 25,
                 25, 26, 27, 27, 28, 29, 29, 30, 31, 32, 32, 33, 34, 35, 35, 36,
                 37, 38, 39, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 50,
                 51, 52, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 66, 67, 68,
                 69, 70, 72, 73, 74, 75, 77, 78, 79, 81, 82, 83, 85, 86, 87, 89,
                 90, 92, 93, 95, 96, 98, 99, 101, 102, 104, 105, 107, 109, 110, 112, 114,
                 115, 117, 119, 120, 122, 124, 126, 127, 129, 131, 133, 135, 137, 138, 140, 142,
                 144, 146, 148, 150, 152, 154, 156, 158, 160, 162, 164, 167, 169, 171, 173, 175,
                 177, 180, 182, 184, 186, 189, 191, 193, 196, 198, 200, 203, 205, 208, 210, 213,
                 215, 218, 220, 223, 225, 228, 231, 233, 236, 239, 241, 244, 247, 249, 252, 255]

        # Assign the color correction profile channel lookup table values
        default_profile.assign_channel_lookup_table_values(0, table)
        default_profile.assign_channel_lookup_table_values(1, table)
        default_profile.assign_channel_lookup_table_values(2, table)

        return default_profile



"""Test led groups."""
from mpf.core.rgb_color import RGBColor
from mpf.tests.MpfTestCase import MpfTestCase


class TestLightGroups(MpfTestCase):

    def get_config_file(self):
        return 'light_groups.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/light/'

    def test_color(self):
        self.machine.light_stripes['stripe1'].color(RGBColor("red"))
        self.advance_time_and_run(1)
        self.assertLightColor("stripe1_light_0", "red")
        self.assertLightColor("stripe1_light_1", "red")
        self.assertLightColor("stripe1_light_2", "red")
        self.assertLightColor("stripe1_light_3", "red")
        self.assertLightColor("stripe1_light_4", "red")

    def test_config(self):
        # stripe 1
        self.assertEqual("led-10-r", self.machine.lights["stripe1_light_0"].hw_drivers["red"][0].number)
        self.assertListEqual(["test", "stripe1"], self.machine.lights["stripe1_light_0"].config['tags'])
        self.assertEqual("led-11-r", self.machine.lights["stripe1_light_1"].hw_drivers["red"][0].number)
        self.assertEqual("led-12-r", self.machine.lights["stripe1_light_2"].hw_drivers["red"][0].number)
        self.assertEqual("led-13-r", self.machine.lights["stripe1_light_3"].hw_drivers["red"][0].number)
        self.assertEqual("led-14-r", self.machine.lights["stripe1_light_4"].hw_drivers["red"][0].number)
        self.assertListEqual(["test", "stripe1"], self.machine.lights["stripe1_light_4"].config['tags'])

        # stripe 2
        self.assertEqual("led-7-200-r", self.machine.lights["stripe2_light_0"].hw_drivers["red"][0].number)
        self.assertEqual(10, self.machine.lights["stripe2_light_0"].config['x'])
        self.assertEqual(20, self.machine.lights["stripe2_light_0"].config['y'])
        self.assertEqual("led-7-201-r", self.machine.lights["stripe2_light_1"].hw_drivers["red"][0].number)
        self.assertEqual(15, self.machine.lights["stripe2_light_1"].config['x'])
        self.assertEqual(20, self.machine.lights["stripe2_light_1"].config['y'])

        # stripe 3
        self.assertEqual("led-ABC-123", self.machine.lights["stripe3_light_0"].hw_drivers["red"][0].number)
        self.assertEqual("led-led-ABC-123+1", self.machine.lights["stripe3_light_0"].hw_drivers["green"][0].number)
        self.assertEqual("led-led-led-ABC-123+1+1",
                         self.machine.lights["stripe3_light_0"].hw_drivers["blue"][0].number)
        self.assertEqual("led-led-led-led-ABC-123+1+1+1",
                         self.machine.lights["stripe3_light_0"].hw_drivers["white"][0].number)
        self.assertEqual("led-led-led-led-led-ABC-123+1+1+1+1",
                         self.machine.lights["stripe3_light_1"].hw_drivers["red"][0].number)

        # ring 1
        self.assertEqual("led-20-r", self.machine.lights["ring1_light_0"].hw_drivers["red"][0].number)
        self.assertEqual("led-21-r", self.machine.lights["ring1_light_1"].hw_drivers["red"][0].number)
        self.assertEqual("led-22-r", self.machine.lights["ring1_light_2"].hw_drivers["red"][0].number)
        self.assertEqual("led-23-r", self.machine.lights["ring1_light_3"].hw_drivers["red"][0].number)
        self.assertEqual("led-24-r", self.machine.lights["ring1_light_4"].hw_drivers["red"][0].number)
        # 90 degree
        self.assertEqual(103, self.machine.lights["ring1_light_0"].config['x'])
        self.assertEqual(50, self.machine.lights["ring1_light_0"].config['y'])
        # 180 degree
        self.assertEqual(100, self.machine.lights["ring1_light_3"].config['x'])
        self.assertEqual(47, self.machine.lights["ring1_light_3"].config['y'])
        # 270 degree
        self.assertEqual(97, self.machine.lights["ring1_light_6"].config['x'])
        self.assertEqual(50, self.machine.lights["ring1_light_6"].config['y'])
        # 360/0 degree
        self.assertEqual(100, self.machine.lights["ring1_light_9"].config['x'])
        self.assertEqual(53, self.machine.lights["ring1_light_9"].config['y'])



"""Contains the MachineVariables class."""

import copy
from platform import platform, python_version, system, release, version, system_alias, machine as platform_machine
from typing import Any, Dict, Optional

from mpf._version import version as mpf_version, extended_version as mpf_extended_version
from mpf.core.data_manager import DataManager
from mpf.core.logging import LogMixin
from mpf.core.utility_functions import Util

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


class MachineVariables(LogMixin):

    """Class for Machine Variables."""

    __slots__ = ["machine", "machine_vars", "machine_var_monitor", "machine_var_data_manager"]

    def __init__(self, machine) -> None:
        """Initialize machine controller."""
        super().__init__()

        self.machine = machine              # type: MachineController
        self.machine_vars = dict()          # type: Dict[str, Any]
        self.machine_var_monitor = False
        self.machine_var_data_manager = None    # type: Optional[DataManager]
        self.configure_logging("machine_vars", self.machine.config['logging']['console']['machine_vars'],
                               self.machine.config['logging']['file']['machine_vars'])

    def load_machine_vars(self, machine_var_data_manager: DataManager, current_time) -> None:
        """Load machine vars from data manager."""
        self.machine_var_data_manager = machine_var_data_manager

        for name, settings in (
                iter(self.machine_var_data_manager.get_data().items())):

            if not isinstance(settings, dict) or "value" not in settings:
                continue

            if ('expire' in settings and settings['expire'] and
                    settings['expire'] < current_time):

                continue

            self.set_machine_var(name=name, value=settings['value'])

        self._load_initial_machine_vars()

        # Create basic system information machine variables
        self.set_machine_var(name="mpf_version", value=mpf_version)
        '''machine_var: mpf_version

        desc: Full version string for MPF.
        '''
        self.set_machine_var(name="mpf_extended_version", value=mpf_extended_version)
        '''machine_var: mpf_extended_version

        desc: Extended version string for MPF. Contains show and bcp version as well.
        '''
        self.set_machine_var(name="python_version", value=python_version())
        '''machine_var: python_version

        desc: Python version.
        '''
        self.set_machine_var(name="platform", value=platform(aliased=True))
        '''machine_var: platform

        desc: A single string identifying the underlying platform
              with as much useful information as possible.
        '''
        platform_info = system_alias(system(), release(), version())
        self.set_machine_var(name="platform_system", value=platform_info[0])
        '''machine_var: platform_system

        desc: Your system (Linux/Windows/Mac).
        '''
        self.set_machine_var(name="platform_release", value=platform_info[1])
        '''machine_var: platform_release

        desc: Release of your operating system.
        '''
        self.set_machine_var(name="platform_version", value=platform_info[2])
        '''machine_var: platform_version

        desc: Version of your operating system.
        '''
        self.set_machine_var(name="platform_machine", value=platform_machine())
        '''machine_var: platform_machine

        desc: Architecture of your machine (32bit/64bit).
        '''

    def __getitem__(self, key):
        """Allow the user to access a machine variable with []. This would be used is machine.variables["var_name"]."""
        return self.get_machine_var(key)

    def __setitem__(self, key, value):
        """Allow the user to set a machine variable with []. Used as machine.variables["var_name"] = value."""
        return self.set_machine_var(key, value)

    def get(self, key):
        """Allow the user to get a machine variable with .get ."""
        return self.get_machine_var(key)

    def _load_initial_machine_vars(self) -> None:
        """Load initial machine var values from config if they did not get loaded from data."""
        if 'machine_vars' not in self.machine.config:
            return

        config = self.machine.config['machine_vars']
        for name, element in config.items():
            if name not in self.machine_vars:
                element = self.machine.config_validator.validate_config("machine_vars", copy.deepcopy(element))
                self.set_machine_var(name=name,
                                     value=Util.convert_to_type(element['initial_value'], element['value_type']))
            self.configure_machine_var(name=name, persist=element.get('persist', False))

    def _write_machine_var_to_disk(self, name: str) -> None:
        """Write value to disk."""
        if self.machine_vars[name]['persist'] and self.machine.config['mpf']['save_machine_vars_to_disk']:
            self._write_machine_vars_to_disk()

    def _write_machine_vars_to_disk(self):
        """Update machine vars on disk."""
        self.machine_var_data_manager.save_all(
            {name: {"value": var["value"], "expire": var['expire_secs']}
             for name, var in self.machine_vars.items() if var["persist"]})

    def get_machine_var(self, name: str) -> Any:
        """Return the value of the variable if it exists, or None if the variable does not exist.

        Args:
        ----
            name: String name of the variable you want to get that value for.
        """
        try:
            return self.machine_vars[name]['value']
        except KeyError:
            return None

    def is_machine_var(self, name: str) -> bool:
        """Return true if machine variable exists."""
        return name in self.machine_vars

    def configure_machine_var(self, name: str, persist: bool, expire_secs: int = None) -> None:
        """Create a new machine variable.

        Args:
        ----
            name: String name of the variable.
            persist: Boolean as to whether this variable should be saved to
                disk so it's available the next time MPF boots.
            expire_secs: Optional number of seconds you'd like this variable
                to persist on disk for. When MPF boots, if the expiration time
                of the variable is in the past, it will not be loaded.
                For example, this lets you write the number of credits on
                the machine to disk to persist even during power off, but you
                could set it so that those only stay persisted for an hour.
        """
        if name not in self.machine_vars:
            self.machine_vars[name] = {'value': None, 'persist': persist, 'expire_secs': expire_secs}
        else:
            self.machine_vars[name]['persist'] = persist
            self.machine_vars[name]['expire_secs'] = expire_secs

    def set_machine_var(self, name: str, value: Any) -> None:
        """Set the value of a machine variable.

        Args:
        ----
            name: String name of the variable you're setting the value for.
            value: The value you're setting. This can be any Type.
        """
        if name not in self.machine_vars:
            self.configure_machine_var(name=name, persist=False)
            prev_value = None
            change = True
        else:
            prev_value = self.machine_vars[name]['value']
            try:
                change = value - prev_value
            except TypeError:
                change = prev_value != value

        # set value
        self.machine_vars[name]['value'] = value

        if change:
            self._write_machine_var_to_disk(name)

            self.debug_log("Setting machine_var '%s' to: %s, (prior: %s, "
                           "change: %s)", name, value, prev_value,
                           change)
            self.machine.events.post('machine_var_' + name,
                                     value=value,
                                     prev_value=prev_value,
                                     change=change)
            '''event: machine_var_(name)
            config_section: machine_vars
            class_label: machine_var

            desc: Posted when a machine variable is added or changes value.
            (Machine variables are like player variables, except they're
            maintained machine-wide instead of per-player or per-game.)

            args:

            value: The new value of this machine variable.

            prev_value: The previous value of this machine variable, e.g. what
            it was before the current value.

            change: If the machine variable just changed, this will be the
            amount of the change. If it's not possible to determine a numeric
            change (for example, if this machine variable is a list), then this
            *change* value will be set to the boolean *True*.
            '''

            if self.machine_var_monitor:
                for callback in self.machine.monitors['machine_vars']:
                    callback(name=name, value=value,
                             prev_value=prev_value, change=change)

    def remove_machine_var(self, name: str) -> None:
        """Remove a machine variable by name.

        If this variable persists to disk, it will remove it from there too.

        Args:
        ----
            name: String name of the variable you want to remove.
        """
        try:
            prev_value = self.machine_vars[name]
            del self.machine_vars[name]
            self._write_machine_vars_to_disk()
        except KeyError:
            pass
        else:
            if self.machine_var_monitor:
                for callback in self.machine.monitors['machine_vars']:
                    callback(name=name, value=None,
                             prev_value=prev_value, change=True)

    def remove_machine_var_search(self, startswith: str = '', endswith: str = '') -> None:
        """Remove a machine variable by matching parts of its name.

        Args:
        ----
            startswith: Optional start of the variable name to match.
            endswith: Optional end of the variable name to match.

        For example, if you pass startswit='player' and endswith='score', this
        method will match and remove player1_score, player2_score, etc.
        """
        for var in list(self.machine_vars.keys()):
            if var.startswith(startswith) and var.endswith(endswith):
                del self.machine_vars[var]

        self._write_machine_vars_to_disk()



"""Test the bonus mode."""
from unittest.mock import MagicMock

from mpf.tests.MpfTestCase import MpfTestCase
from mpf._version import version, extended_version


class TestMachineVariables(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/machine_vars/'

    def _get_mock_data(self):
        return {"machine_vars": {"player2_score": {"value": 118208660},
                                 "player3_score": {"value": 17789290},
                                 "player4_score": {"value": 3006600},
                                 "another_score": {"value": 123},
                                 "expired_value": {"value": 23, "expire": self.clock.get_time() - 100},
                                 "not_expired_value": {"value": 24, "expire": self.clock.get_time() + 100},
                                 "test1": {"value": 42}},
                }

    def testSystemInfoVariables(self):
        self.assertTrue(self.machine.variables.is_machine_var("mpf_version"))
        self.assertTrue(self.machine.variables.is_machine_var("mpf_extended_version"))
        self.assertTrue(self.machine.variables.is_machine_var("python_version"))
        self.assertTrue(self.machine.variables.is_machine_var("platform"))
        self.assertTrue(self.machine.variables.is_machine_var("platform_system"))
        self.assertTrue(self.machine.variables.is_machine_var("platform_release"))
        self.assertTrue(self.machine.variables.is_machine_var("platform_version"))
        self.assertTrue(self.machine.variables.is_machine_var("platform_machine"))
        self.assertEqual(version, self.machine.variables.get_machine_var("mpf_version"))
        self.assertEqual(extended_version, self.machine.variables.get_machine_var("mpf_extended_version"))

    def testVarLoadAndRemove(self):
        self.assertFalse(self.machine.variables.is_machine_var("expired_value"))
        self.assertTrue(self.machine.variables.is_machine_var("not_expired_value"))
        self.assertTrue(self.machine.variables.is_machine_var("player2_score"))
        # should always persist
        #self.assertTrue(self.machine.variables.machine_vars["player2_score"]["persist"])
        # random variable does not persist
        self.assertFalse(self.machine.variables.machine_vars["another_score"]["persist"])
        # configured to persist
        self.assertTrue(self.machine.variables.machine_vars["test1"]["persist"])
        self.assertTrue(self.machine.variables.machine_vars["test2"]["persist"])
        # configured to not persist
        self.assertFalse(self.machine.variables.machine_vars["test3"]["persist"])
        self.assertEqual(118208660, self.machine.variables.get_machine_var("player2_score"))

        self.machine.variables.remove_machine_var("player2_score")

        self.assertFalse(self.machine.variables.is_machine_var("player2_score"))
        self.assertTrue(self.machine.variables.is_machine_var("player3_score"))

        self.machine.variables.machine_var_data_manager._trigger_save = MagicMock()
        self.machine.variables.remove_machine_var_search(startswith="player", endswith="_score")
        self.assertFalse(self.machine.variables.is_machine_var("player2_score"))
        self.assertFalse(self.machine.variables.is_machine_var("player3_score"))

        self.assertEqual(123, self.machine.variables.get_machine_var("another_score"))

        self.advance_time_and_run(10)

        self.machine.variables.machine_var_data_manager._trigger_save.assert_called_with()
        self.assertEqual({'test1': {'value': 42, 'expire': None}, 'test2': {'value': '5', 'expire': None}},
                         self.machine.variables.machine_var_data_manager.data)

    def testVarSetAndGet(self):
        self.assertEqual(118208660, self.machine.variables["player2_score"])

        self.assertEqual(None, self.machine.variables["player1_score"])
        self.machine.variables["player1_score"] = 123
        self.assertEqual(123, self.machine.variables["player1_score"])

        self.assertEqual(17789290, self.machine.variables.get("player3_score"))


class TestMalformedMachineVariables(MpfTestCase):

    def _get_mock_data(self):
        return {"machine_vars": {"player2_score": {"value": 118208660},
                                 "player3_score": {"value": 17789290},
                                 "player4_score": {"value": 3006600},
                                 "player5_score": 123,
                                 "player6_score": {"asd": 3006600},
                                 "value": 0}}

    def testVarLoads(self):
        self.assertTrue(self.machine.variables.is_machine_var("player2_score"))
        self.assertEqual(118208660, self.machine.variables.get_machine_var("player2_score"))
        self.assertFalse(self.machine.variables.is_machine_var("player5_score"))
        self.assertEqual(None, self.machine.variables.get_machine_var("player5_score"))




"""Variable Config Player (used for scoring and more)."""
import re
from collections import namedtuple
from typing import Dict, List, Any

from mpf.core.config_player import ConfigPlayer
from mpf.core.machine import MachineController

VarBlock = namedtuple("VarBlock", ["priority", "context"])


class VariablePlayer(ConfigPlayer):

    """Posts events based on config."""

    config_file_section = 'variable_player'
    show_section = 'variables'

    __slots__ = ["blocks"]

    def __init__(self, machine: MachineController) -> None:
        """Initialise variable player."""
        super().__init__(machine)
        self.blocks = {}    # type: Dict[str, List[VarBlock]]

    @staticmethod
    def is_entry_valid_outside_mode(settings: dict) -> bool:
        """Return true if this entry may run without a game and player."""
        for event, setting in settings.items():
            del event
            if setting['action'] not in ("set_machine", "add_machine"):
                return False

        # true if only set_machine or add_machine are used
        return True

    # pylint: disable-msg=too-many-arguments
    def handle_subscription_change(self, value, settings, priority, context, key):
        """Handle subscriptions."""
        for var, s in settings.items():
            if var == "block":
                self.raise_config_error('Do not use "block" as variable name in variable_player.', 1, context=context)

            if s['action'] not in ("set", "set_machine"):
                self.raise_config_error('Cannot use add on subscriptions. '
                                        'Use action set or set_machine.', 8, context=context)

            args = {"value": value}

            if s['condition'] and not s['condition'].evaluate(args):
                continue

            block_item = var + ":" + str(key)

            if self._is_blocked(block_item, context, priority):
                continue
            if s['block']:
                if block_item not in self.blocks:
                    self.blocks[block_item] = []
                if VarBlock(priority, context) not in self.blocks[block_item]:
                    self.blocks[block_item].append(VarBlock(priority, context))

            self._set_variable(var, s, args, context)

    def play(self, settings: dict, context: str, calling_context: str,
             priority: int = 0, **kwargs) -> None:
        """Variable name."""
        for var, s in settings.items():
            if var == "block":
                self.raise_config_error('Do not use "block" as variable name in variable_player.', 1, context=context)

            if s['action'] in ("add", "add_machine") and s['string']:
                self.raise_config_error('Cannot add two strings. Use action set or set_machine.', 3, context=context)

            if s['condition'] and not s['condition'].evaluate(kwargs):
                continue

            block_item = var + ":" + str(calling_context)

            if self._is_blocked(block_item, context, priority):
                continue
            if s['block']:
                if block_item not in self.blocks:
                    self.blocks[block_item] = []
                if VarBlock(priority, context) not in self.blocks[block_item]:
                    self.blocks[block_item].append(VarBlock(priority, context))

            self._set_variable(var, s, kwargs, context)

    def _is_blocked(self, block_item: str, context: str,
                    priority: int) -> bool:
        if block_item not in self.blocks or not self.blocks[block_item]:
            return False
        priority_sorted = sorted(self.blocks[block_item], reverse=True)
        first_element = priority_sorted[0]
        return first_element.priority > priority and first_element.context != context

    def _set_variable(self, var: str, entry: dict, placeholder_parameters: dict, context) -> None:
        # evaluate placeholder
        if entry['float']:
            value = entry['float'].evaluate(placeholder_parameters)
        elif entry['int']:
            value = entry['int'].evaluate(placeholder_parameters)
        elif entry['string']:
            value = entry['string'].evaluate(placeholder_parameters)
        else:
            value = None    # prevent type confusion
            self.raise_config_error("You need to either set float, int or string", 2, context=context)

        if entry['action'] == "add":
            assert self.machine.game is not None
            assert self.machine.game.player is not None
            if entry['player']:
                # specific player
                try:
                    self.machine.game.player_list[entry['player'] - 1][var] += value
                except IndexError:
                    self.warning_log("Failed to set player var %s for player %s. There are only %s players.",
                                     var, entry['player'] - 1, self.machine.game.num_players)
            else:
                # default to current player
                self.machine.game.player[var] += value
        elif entry['action'] == "set":
            assert self.machine.game is not None
            assert self.machine.game.player is not None
            if entry['player']:
                # specific player
                try:
                    self.machine.game.player_list[entry['player'] - 1][var] = value
                except IndexError:
                    self.warning_log("Failed to set player var %s for player %s. There are only %s players.",
                                     var, entry['player'] - 1, self.machine.game.num_players)
            else:
                # default to current player
                self.machine.game.player[var] = value
        elif entry['action'] == "add_machine":
            old_value = self.machine.variables.get_machine_var(var)
            if old_value is None:
                old_value = 0
            self.machine.variables.set_machine_var(var, old_value + value)
        elif entry['action'] == "set_machine":
            self.machine.variables.set_machine_var(var, value)
        else:
            self.raise_config_error("Invalid value: {}".format(entry), 8, context=context)

    def clear_context(self, context: str) -> None:
        """Clear context."""
        for var in self.blocks:
            for entry, s in enumerate(self.blocks[var]):
                if s.context == context:
                    del self.blocks[var][entry]

    def validate_config_entry(self, settings: dict, name: str) -> dict:
        """Validate one entry of this player."""
        config = {}
        if not isinstance(settings, dict):
            self.raise_config_error("Settings of variable_player should "
                                    "be a dict. But are: {}".format(settings), 5, context=name)
        for var, s in settings.items():
            var_conditional_event = self.machine.placeholder_manager.parse_conditional_template(var)
            value_dict = self._parse_config(s, name)
            value_dict["condition"] = var_conditional_event.condition
            config[var_conditional_event.name] = value_dict
            if not bool(re.match('^[0-9a-zA-Z_-]+$', var_conditional_event.name)):
                self.raise_config_error("Variable may only contain letters, numbers, dashes and underscores. "
                                        "Name: {}".format(var_conditional_event.name), 4, context=name)
        return config

    def get_express_config(self, value: Any) -> dict:
        """Parse express config."""
        if not isinstance(value, str):
            block = False
        else:
            try:
                value, block_str = value.split('|')
            except ValueError:
                block = False
            else:
                if block_str != "block":
                    self.raise_config_error("Invalid action in variable_player entry: {}".format(value), 6)
                block = True

        return {"int": value, "block": block}

    def get_list_config(self, value: Any):
        """Parse list."""
        self.raise_config_error("Variable player does not support lists.", 7)



"""Switch ball counter."""
import asyncio

from mpf.core.utility_functions import Util
from mpf.devices.ball_device.physical_ball_counter import PhysicalBallCounter, BallLostActivity, \
    BallEntranceActivity, UnknownBallActivity, BallReturnActivity


class SwitchCounter(PhysicalBallCounter):

    """Determine ball count by counting switches.

    This should be used for devices with multiple switches and/or a jam switch. Simple devices with only one switch
    should use a simpler counter.
    """

    __slots__ = ["_entrances", "_trigger_recount", "_task", "_is_unreliable", "_switches"]

    def __init__(self, ball_device, config):
        """Initialise ball counter."""
        for option in ["entrance_count_delay", "exit_count_delay", "entrance_event_timeout", "jam_switch",
                       "ball_switches"]:
            if option not in config and option in ball_device.config:
                config[option] = ball_device.config[option]
        super().__init__(ball_device, config)

        self.config = self.machine.config_validator.validate_config("ball_device_counter_ball_switches", self.config)
        self._entrances = []
        self._switches = set(self.config['ball_switches'])
        if self.config['jam_switch']:
            self._switches.add(self.config['jam_switch'])
        self._trigger_recount = asyncio.Event()
        # Register switch handlers with delays for entrance & exit counts
        for switch in self._switches:
            self.machine.switch_controller.add_switch_handler_obj(
                switch=switch, state=1,
                ms=self.config['entrance_count_delay'],
                callback=self.trigger_recount)
            self.machine.switch_controller.add_switch_handler_obj(
                switch=switch, state=1,
                callback=self.invalidate_count)
            self.machine.switch_controller.add_switch_handler_obj(
                switch=switch, state=0,
                ms=self.config['exit_count_delay'],
                callback=self.trigger_recount)
            self.machine.switch_controller.add_switch_handler_obj(
                switch=switch, state=0,
                callback=self.invalidate_count)

        self._task = self.machine.clock.loop.create_task(self._run())
        self._task.add_done_callback(Util.raise_exceptions)
        self._is_unreliable = False

        # make sure timeouts are reasonable:
        # exit_count_delay < all eject_timeout
        if self.config['exit_count_delay'] > min(
                self.ball_device.config['eject_timeouts'].values()):
            self.ball_device.raise_config_error('Configuration error in {} ball device. '
                                                'all eject_timeouts have to be larger than '
                                                'exit_count_delay'.
                                                format(self.ball_device.name), 6)

        # entrance_count_delay < all eject_timeout
        if self.config['entrance_count_delay'] > min(
                self.ball_device.config['eject_timeouts'].values()):
            self.ball_device.raise_config_error('Configuration error in {} ball device. '
                                                'all eject_timeouts have to be larger than '
                                                'entrance_count_delay'.
                                                format(self.ball_device.name), 7)

        # multiple switches + mechanical eject is not supported
        if (self.capacity > 1 and
                self.ball_device.config['mechanical_eject']):
            self.ball_device.raise_config_error('Configuration error in {} ball device. '
                                                'mechanical_eject can only be used with '
                                                'devices that have 1 ball switch'.
                                                format(self.ball_device.name), 5)

        for switch in self._switches:
            if switch and '{}_active'.format(self.ball_device.config['captures_from'].name) in switch.tags:
                self.ball_device.raise_config_error(
                    "Ball device '{}' uses switch '{}' which has a "
                    "'{}_active' tag. This is handled internally by the device. Remove the "
                    "redundant '{}_active' tag from that switch.".format(
                        self.ball_device.name, switch.name, self.ball_device.config['captures_from'].name,
                        self.ball_device.config['captures_from'].name), 13)

        # cannot have ball switches and capacity
        if self.ball_device.config.get('ball_capacity'):
            self.ball_device.raise_config_error("Cannot use capacity and ball switches.", 3)

    def stop(self):
        """Stop task."""
        super().stop()
        if self._task:
            self._task.cancel()

    def trigger_recount(self):
        """Trigger a count."""
        self._trigger_recount.set()

    async def _recount(self):
        while True:
            await self._trigger_recount.wait()
            self._trigger_recount.clear()
            try:
                balls = self.count_balls_sync()
                return balls
            except ValueError:
                continue

    async def _run(self):
        self._trigger_recount.set()
        while True:
            new_count = await self._recount()
            self._count_stable.set()

            if self._last_count is None:
                self._last_count = new_count
            elif self._last_count < 0:
                raise AssertionError("Count may never be negative")

            if self.is_jammed() and new_count == 1:
                # only jam is active. keep previous count
                self.debug_log("Counter is jammed. Only jam is active. Will no longer trust count.")
                # ball potentially returned
                if not self._is_unreliable:
                    self.trigger_activity()
                    self.record_activity(BallReturnActivity())
                    self._is_unreliable = True
                continue

            self._is_unreliable = False

            if new_count == self._last_count:
                # count did not change
                continue

            if new_count > self._last_count:
                # new ball
                for _ in range(new_count - self._last_count):
                    try:
                        last_entrance = self._entrances.pop(0)
                    except IndexError:
                        last_entrance = -1000

                    if last_entrance > self.machine.clock.get_time() - self.config['entrance_event_timeout']:
                        self.record_activity(BallEntranceActivity())
                    else:
                        self.record_activity(UnknownBallActivity())
            elif new_count < self._last_count:
                # lost ball
                for _ in range(self._last_count - new_count):
                    self.record_activity(BallLostActivity())

            # update count
            self._last_count = new_count
            self.trigger_activity()

    def received_entrance_event(self):
        """Handle entrance event."""
        entrance_time = self.machine.clock.get_time()
        entrance_timeout = self.config['entrance_event_timeout']
        self._entrances = [entrance for entrance in self._entrances if entrance > entrance_time - entrance_timeout]
        self._entrances.append(entrance_time)

    def _count_switches_sync(self):
        """Return active switches or raise ValueError if switches are unstable."""
        switches = []
        for switch in self._switches:
            valid = False
            if self.machine.switch_controller.is_active(
                    switch, ms=self.config['entrance_count_delay']):
                switches.append(switch)
                valid = True
            elif self.machine.switch_controller.is_inactive(
                    switch, ms=self.config['exit_count_delay']):
                valid = True

            if not valid:
                # one of our switches wasn't valid long enough
                self.debug_log("Switch '%s' changed too recently. Aborting count!", switch.name)
                raise ValueError('Count not stable yet. Run again!')

        return switches

    def count_balls_sync(self):
        """Count currently active switches or raise ValueError if switches are unstable."""
        switches = self._count_switches_sync()
        ball_count = len(switches)

        self.debug_log("Counted %s balls. Active switches: %s. Old: %s", ball_count, switches, self._last_count)
        return ball_count

    @property
    def capacity(self):
        """Return capacity under normal circumstances (i.e. without jam switches)."""
        return len(self.config['ball_switches'])

    def is_jammed(self):
        """Return true if the jam switch is currently active."""
        return self.config['jam_switch'] and self.machine.switch_controller.is_active(
            self.config['jam_switch'], ms=self.config['entrance_count_delay'])

    def is_count_unreliable(self):
        """Return true if only the jam switch is active and the count is unknown."""
        return self._is_unreliable

    @property
    def is_ready_to_receive(self):
        """Return true if count is stable and we got at least one slot."""
        try:
            count = len(self._count_switches_sync())
        except ValueError:
            # count not stable
            return False
        # we intentionally do not consider jam_switch here (which would be part of self._switches)
        return count != self.capacity

    async def wait_for_ready_to_receive(self):
        """Wait until there is at least on inactive switch."""
        # future returns when ball_count != number of switches
        # we intentionally do not consider jam_switch here (which would be part of self._switches)
        return await self.wait_for_ball_count_changes(self.capacity)

    async def wait_for_ball_to_leave(self):
        """Wait for any active switch to become inactive."""
        while True:
            waiter = self.wait_for_count_stable()
            try:
                active_switches = self._count_switches_sync()
                waiter.cancel()
                break
            except ValueError:
                await waiter

        waiters = []
        for switch in active_switches:
            waiters.append(self.machine.switch_controller.wait_for_switch(
                switch=switch, state=0))

        if not waiters:
            self.ball_device.log.warning("No switch is active. Cannot wait on empty list.")
            future = asyncio.Future()
            future.set_result(True)
            return future

        done_future = asyncio.ensure_future(Util.first(waiters))
        done_future.add_done_callback(self._ball_left)
        return done_future

    def _ball_left(self, future):
        if future.cancelled():
            return
        self._last_count -= 1
        self.record_activity(BallLostActivity())
        self.trigger_recount()
        self.trigger_activity()



from mpf.tests.MpfTestCase import MpfTestCase


class TestSettings(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/settings/'

    def get_platform(self):
        return 'smart_virtual'

    def _tilted(self, **kwargs):
        del kwargs
        self._is_tilted = True

    def test_settings_values(self):
        self.assertEqual(0, self.machine.settings.custom_setting_int)
        self.machine.settings.set_setting_value("custom_setting_int", 2)
        self.assertEqual(2, self.machine.settings.custom_setting_int)

        self.assertEqual("one", self.machine.settings.custom_setting_str)
        self.machine.settings.set_setting_value("custom_setting_str", "zero")
        self.assertEqual("zero", self.machine.settings.custom_setting_str)



"""Test event manager."""
from mpf.core.delays import DelayManager
from mpf.core.settings_controller import SettingEntry
from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase
from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import patch


class TestEventManager(MpfFakeGameTestCase, MpfTestCase):
    def __init__(self, test_map):
        super().__init__(test_map)
        self._handler1_args = tuple()
        self._handler1_kwargs = dict()
        self._handler1_called = 0
        self._handler2_args = tuple()
        self._handler2_kwargs = dict()
        self._handler2_called = 0
        self._handler3_args = tuple()
        self._handler3_kwargs = dict()
        self._handler3_called = 0
        self._handlers_called = list()
        self._handler_returns_false_args = tuple()
        self._handler_returns_false_kwargs = dict()
        self._handler_returns_false_called = 0
        self._relay1_called = 0
        self._relay2_called = 0
        self._relay_callback_args = tuple()
        self._relay_callback_kwargs = dict()
        self._relay_callback_called = 0
        self._callback_args = tuple()
        self._callback_kwargs = dict()
        self._callback_called = 0
        self._queue = None
        self._queue_callback_args = tuple()
        self._queue_callback_kwargs = dict()
        self._queue_callback_called = 0

    def get_config_file(self):
        return 'test_event_manager.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/event_manager/'

    def event_handler1(self, *args, **kwargs):
        self._handler1_args = args
        self._handler1_kwargs = kwargs
        self._handler1_called += 1
        self._handlers_called.append(self.event_handler1)

    def event_handler2(self, *args, **kwargs):
        self._handler2_args = args
        self._handler2_kwargs = kwargs
        self._handler2_called += 1
        self._handlers_called.append(self.event_handler2)

    def event_handler3(self, *args, **kwargs):
        self._handler3_args = args
        self._handler3_kwargs = kwargs
        self._handler3_called += 1
        self._handlers_called.append(self.event_handler3)

    def event_handler_returns_false(self, *args, **kwargs):
        self._handler_returns_false_args = args
        self._handler_returns_false_kwargs = kwargs
        self._handler_returns_false_called += 1
        self._handlers_called.append(self.event_handler_returns_false)

        return False

    def event_handler_relay1(self, relay_test, **kwargs):
        del kwargs
        self._relay1_called += 1
        self._handlers_called.append(self.event_handler_relay1)

        return {'relay_test': relay_test}

    def event_handler_relay2(self, relay_test, **kwargs):
        del kwargs
        self._relay2_called += 1
        self._handlers_called.append(self.event_handler_relay2)

        return {'relay_test': relay_test - 1}

    def callback(self, *args, **kwargs):
        self._callback_args = args
        self._callback_kwargs = kwargs
        self._callback_called += 1
        self._handlers_called.append(self.callback)

    def relay_callback(self, *args, **kwargs):
        self._relay_callback_args = args
        self._relay_callback_kwargs = kwargs
        self._relay_callback_called += 1
        self._handlers_called.append(self.relay_callback)

    def event_handler_calls_second_event(self, **kwargs):
        del kwargs
        self.machine.events.post('second_event')
        self._handlers_called.append(self.event_handler_calls_second_event)

    def event_handler_add_queue(self, queue, **kwargs):
        del kwargs
        self._handlers_called.append(self.event_handler_add_queue)
        self._queue = queue
        self._queue.wait()

    def event_handler_add_quick_queue(self, queue, **kwargs):
        del kwargs
        self._handlers_called.append(self.event_handler_add_quick_queue)
        self._queue = queue
        self._queue.wait()
        self._queue.clear()

    def event_handler_clear_queue(self, **kwargs):
        del kwargs
        self._handlers_called.append(self.event_handler_clear_queue)
        self._queue.clear()

    def queue_callback(self, *args, **kwargs):
        self._queue_callback_args = args
        self._queue_callback_kwargs = kwargs
        self._queue_callback_called += 1
        self._handlers_called.append(self.queue_callback)

    def test_event(self):
        # tests that a handler responds to a regular event post
        self.machine.events.add_handler('test_event', self.event_handler1)
        self.advance_time_and_run(1)

        self.machine.events.post('test_event')
        self.advance_time_and_run(1)

        self.assertEqual(1, self._handler1_called)
        self.assertEqual(tuple(), self._handler1_args)
        self.assertEqual(dict(), self._handler1_kwargs)

    def test_event_double_register(self):
        # tests that a handler responds to a regular event post
        self.machine.events.add_handler('test_event', self.event_handler1)
        self.machine.events.add_handler('test_event', self.event_handler1)
        self.advance_time_and_run(1)

        self.machine.events.post('test_event')
        self.advance_time_and_run(1)

        self.assertEqual(2, self._handler1_called)
        self.assertEqual(tuple(), self._handler1_args)
        self.assertEqual(dict(), self._handler1_kwargs)

    def test_event_with_kwargs(self):
        # test that a kwarg can be passed to a handler which is registered for
        # a regular event post
        self.machine.events.add_handler('test_event', self.event_handler1)
        self.advance_time_and_run(1)

        self.machine.events.post('test_event', test1='test1')
        self.advance_time_and_run(1)

        self.assertEqual(1, self._handler1_called)
        self.assertEqual(tuple(), self._handler1_args)
        self.assertEqual({'test1': 'test1'}, self._handler1_kwargs)

    def test_event_with_callback(self):
        # test that a callback is called when the event is done
        self.machine.events.add_handler('test_event', self.event_handler1)
        self.advance_time_and_run(1)

        self.machine.events.post('test_event', test1='test1',
                                 callback=self.callback)
        self.advance_time_and_run(1)

        self.assertEqual(1, self._callback_called)

    def test_nested_callbacks(self):
        # tests that an event handlers which posts another event has that event
        # handled before the first event's callback is called

        self.machine.events.add_handler('test_event',
                                        self.event_handler_calls_second_event)
        self.machine.events.add_handler('second_event', self.event_handler1)

        self.advance_time_and_run(1)

        self.machine.events.post('test_event', callback=self.callback)
        self.advance_time_and_run(1)

        self.assertEqual(self._handlers_called[0],
                         self.event_handler_calls_second_event)
        self.assertEqual(self._handlers_called[1],
                         self.event_handler1)
        self.assertEqual(self._handlers_called[2],
                         self.callback)

    def test_event_handler_priorities(self):
        # tests that handler priorities work. The second handler should be
        # called first because it's a higher priority even though it's
        # registered second
        self.machine.events.add_handler('test_event', self.event_handler1,
                                        priority=100)
        self.machine.events.add_handler('test_event', self.event_handler2,
                                        priority=200)
        self.advance_time_and_run(1)

        self.machine.events.post('test_event')
        self.advance_time_and_run(1)

        self.assertEqual(1, self._handler1_called)
        self.assertEqual(tuple(), self._handler1_args)
        self.assertEqual(dict(), self._handler1_kwargs)
        self.assertEqual(1, self._handler2_called)
        self.assertEqual(tuple(), self._handler2_args)
        self.assertEqual(dict(), self._handler2_kwargs)

        self.assertEqual(self._handlers_called[0], self.event_handler2)
        self.assertEqual(self._handlers_called[1], self.event_handler1)

    def test_remove_handler_by_handler(self):
        # tests that a handler can be removed by passing the handler to remove
        self.machine.events.add_handler('test_event', self.event_handler1)
        self.advance_time_and_run(1)

        self.machine.events.post('test_event')
        self.advance_time_and_run(1)

        self.assertEqual(1, self._handler1_called)
        self.assertEqual(tuple(), self._handler1_args)
        self.assertEqual(dict(), self._handler1_kwargs)

        self.machine.events.remove_handler(self.event_handler1)

        self.machine.events.post('test_event')
        self.advance_time_and_run(1)

        self.assertEqual(1, self._handler1_called)
        self.assertEqual(tuple(), self._handler1_args)
        self.assertEqual(dict(), self._handler1_kwargs)

    def test_remove_handler_by_event(self):
        # tests that a handler can be removed by a handler/event combo, and
        # that only that handler/event combo is removed
        self.machine.events.add_handler('test_event1', self.event_handler1)
        self.machine.events.add_handler('test_event2', self.event_handler1)
        self.advance_time_and_run(1)

        self.machine.events.post('test_event1')
        self.advance_time_and_run(1)

        self.assertEqual(1, self._handler1_called)
        self.assertEqual(tuple(), self._handler1_args)
        self.assertEqual(dict(), self._handler1_kwargs)

        # should not remove handler since this is the wrong event
        self.machine.events.remove_handler_by_event('test_event3',
                                                    self.event_handler1)

        self.machine.events.post('test_event1')
        self.advance_time_and_run(1)

        self.assertEqual(2, self._handler1_called)
        self.assertEqual(tuple(), self._handler1_args)
        self.assertEqual(dict(), self._handler1_kwargs)

        # remove handler for this event
        self.machine.events.remove_handler_by_event('test_event1',
                                                    self.event_handler1)

        self.machine.events.post('test_event1')
        self.advance_time_and_run(1)

        # results should be the same as above since this handler was removed
        self.assertEqual(2, self._handler1_called)
        self.assertEqual(tuple(), self._handler1_args)
        self.assertEqual(dict(), self._handler1_kwargs)

        self.machine.events.post('test_event2')
        self.advance_time_and_run(1)

        # results should be the same as above since this handler was removed
        self.assertEqual(3, self._handler1_called)
        self.assertEqual(tuple(), self._handler1_args)
        self.assertEqual(dict(), self._handler1_kwargs)

    def test_remove_handler_by_key(self):
        # tests that a handler responds to a regular event post
        key = self.machine.events.add_handler('test_event',
                                              self.event_handler1)
        self.advance_time_and_run(1)

        self.machine.events.post('test_event')
        self.advance_time_and_run(1)

        self.assertEqual(1, self._handler1_called)
        self.assertEqual(tuple(), self._handler1_args)
        self.assertEqual(dict(), self._handler1_kwargs)

        self.machine.events.remove_handler_by_key(key)

        self.machine.events.post('test_event')
        self.advance_time_and_run(1)

        self.assertEqual(1, self._handler1_called)
        self.assertEqual(tuple(), self._handler1_args)
        self.assertEqual(dict(), self._handler1_kwargs)

    def test_remove_handlers_by_keys(self):
        # tests that multiple handlers can be removed by an iterable keys list
        keys = list()

        keys.append(self.machine.events.add_handler('test_event1',
                                                    self.event_handler1))
        keys.append(self.machine.events.add_handler('test_event2',
                                                    self.event_handler2))
        self.machine.events.post('test_event1')
        self.advance_time_and_run(1)

        self.assertEqual(1, self._handler1_called)
        self.assertEqual(tuple(), self._handler1_args)
        self.assertEqual(dict(), self._handler1_kwargs)

        self.machine.events.post('test_event2')
        self.advance_time_and_run(1)

        self.assertEqual(1, self._handler2_called)
        self.assertEqual(tuple(), self._handler2_args)
        self.assertEqual(dict(), self._handler2_kwargs)

        self.machine.events.remove_handlers_by_keys(keys)

        # post events again and handlers should not be called again
        self.machine.events.post('test_event1')
        self.advance_time_and_run(1)

        self.assertEqual(1, self._handler1_called)
        self.assertEqual(tuple(), self._handler1_args)
        self.assertEqual(dict(), self._handler1_kwargs)

        self.machine.events.post('test_event2')
        self.advance_time_and_run(1)

        self.assertEqual(1, self._handler2_called)
        self.assertEqual(tuple(), self._handler2_args)
        self.assertEqual(dict(), self._handler2_kwargs)

    def test_does_event_exist(self):
        self.machine.events.add_handler('test_event', self.event_handler1)

        self.assertEqual(True,
                         self.machine.events.does_event_exist('test_event'))
        self.assertEqual(False,
                         self.machine.events.does_event_exist('test_event1'))

    def test_regular_event_with_false_return(self):
        # tests that regular events process all handlers even if one returns
        # False

        self.machine.events.add_handler('test_event', self.event_handler1,
                                        priority=100)
        self.machine.events.add_handler('test_event',
                                        self.event_handler_returns_false,
                                        priority=200)
        self.advance_time_and_run(1)

        self.machine.events.post('test_event')
        self.advance_time_and_run(1)

        self.assertEqual(1, self._handler1_called)
        self.assertEqual(1, self._handler_returns_false_called)

        self.assertEqual(self._handlers_called[0],
                         self.event_handler_returns_false)
        self.assertEqual(self._handlers_called[1], self.event_handler1)

    def test_post_boolean(self):
        # tests that a boolean event works

        self.machine.events.add_handler('test_event', self.event_handler1,
                                        priority=100)
        self.machine.events.add_handler('test_event', self.event_handler2,
                                        priority=200)
        self.advance_time_and_run(1)

        self.machine.events.post_boolean('test_event')
        self.advance_time_and_run(1)

        self.assertEqual(1, self._handler1_called)
        self.assertEqual(tuple(), self._handler1_args)
        self.assertEqual(dict(), self._handler1_kwargs)
        self.assertEqual(1, self._handler2_called)
        self.assertEqual(tuple(), self._handler2_args)
        self.assertEqual(dict(), self._handler2_kwargs)

        self.assertEqual(self._handlers_called[0], self.event_handler2)
        self.assertEqual(self._handlers_called[1], self.event_handler1)

    def test_boolean_event_with_false_return(self):
        # tests that regular events process all handlers even if one returns
        # False

        self.machine.events.add_handler('test_event', self.event_handler1,
                                        priority=100)
        self.machine.events.add_handler('test_event',
                                        self.event_handler_returns_false,
                                        priority=200)
        self.advance_time_and_run(1)

        self.machine.events.post_boolean('test_event')
        self.advance_time_and_run(1)

        self.assertEqual(0, self._handler1_called)
        self.assertEqual(1, self._handler_returns_false_called)

        self.assertEqual(self._handlers_called[0],
                         self.event_handler_returns_false)

        self.assertEqual(1, len(self._handlers_called))

    def test_relay_event(self):
        # tests that a relay event works by passing a value

        self.machine.events.add_handler('test_event',
                                        self.event_handler_relay1,
                                        priority=200)

        self.advance_time_and_run(1)

        self.machine.events.post_relay('test_event', relay_test=1,
                                       callback=self.relay_callback)
        self.advance_time_and_run(1)

        self.assertEqual(1, self._relay1_called)
        self.assertEqual(1, self._relay_callback_called)

        assert 'relay_test' in self._relay_callback_kwargs
        assert self._relay_callback_kwargs['relay_test'] == 1

    def test_relay_event_handler_changes_value(self):
        # tests that a relay event works by passing a value to a handler that
        # changes that value

        self.machine.events.add_handler('test_event',
                                        self.event_handler_relay1,
                                        priority=200)
        self.machine.events.add_handler('test_event',
                                        self.event_handler_relay2,
                                        priority=100)
        self.advance_time_and_run(1)

        self.machine.events.post_relay('test_event', relay_test=1,
                                       callback=self.relay_callback)
        self.advance_time_and_run(1)

        self.assertEqual(1, self._relay1_called)
        self.assertEqual(1, self._relay2_called)
        self.assertEqual(1, self._relay_callback_called)

        assert 'relay_test' in self._relay_callback_kwargs
        self.assertEqual(self._relay_callback_kwargs['relay_test'], 0)

    def test_queue(self):
        # tests that a queue event works by registering and clearing a queue

        self.machine.events.add_handler('test_event',
                                        self.event_handler_add_queue)

        self.advance_time_and_run(1)

        self.machine.events.post_queue('test_event',
                                       callback=self.queue_callback)
        self.advance_time_and_run(1)

        self.assertEqual(self._handlers_called.count(self.event_handler_add_queue), 1)
        self.assertEqual(self._handlers_called.count(self.queue_callback), 0)
        self.assertEqual(False, self._queue.is_empty())

        self.event_handler_clear_queue()
        self.advance_time_and_run()

        self.assertEqual(self._handlers_called.count(self.queue_callback), 1)
        self.assertEqual(True, self._queue.is_empty())

    def test_queue_event_with_no_queue(self):
        # tests that a queue event works and the callback is called right away
        # if no handlers request a wait

        self.machine.events.add_handler('test_event', self.event_handler1)

        self.advance_time_and_run(1)

        self.machine.events.post_queue('test_event',
                                       callback=self.queue_callback)

        self.advance_time_and_run(1)

        self.assertEqual(self._handlers_called.count(self.event_handler1), 1)
        self.assertEqual(self._handlers_called.count(self.queue_callback), 1)

    def test_queue_event_with_quick_queue_clear(self):
        # tests that a queue event that quickly creates and clears a queue

        self.machine.events.add_handler('test_event',
                                        self.event_handler_add_quick_queue)

        self.advance_time_and_run(1)

        self.machine.events.post_queue('test_event',
                                       callback=self.queue_callback)
        self.advance_time_and_run(1)

        self.assertEqual(self._handlers_called.count(self.event_handler_add_quick_queue), 1)

        self.assertEqual(self._handlers_called.count(self.queue_callback), 1)
        self.assertEqual(True, self._queue.is_empty())

    def test_queue_event_with_no_registered_handlers(self):
        # tests that a queue event callback is called works even if there are
        # not registered handlers for that event

        self.machine.events.post_queue('test_event',
                                       callback=self.queue_callback)
        self.advance_time_and_run(1)

        self.assertEqual(self._handlers_called.count(self.queue_callback), 1)
        self.assertIsNone(self._queue)

    def test_queue_event_with_double_quick_queue_clear(self):
        # tests that a queue event that quickly creates and clears a queue

        self.machine.events.add_handler('test_event',
                                        self.event_handler_add_quick_queue,
                                        priority=1)
        self.machine.events.add_handler('test_event',
                                        self.event_handler_add_quick_queue,
                                        priority=2)

        self.advance_time_and_run(1)

        self.machine.events.post_queue('test_event',
                                       callback=self.queue_callback)
        self.advance_time_and_run(1)

        self.assertEqual(self._handlers_called.count(self.event_handler_add_quick_queue), 2)

        self.assertEqual(self._handlers_called.count(self.queue_callback), 1)
        self.assertEqual(True, self._queue.is_empty())

    def test_event_player(self):
        self.machine.events.add_handler('test_event_player1', self.event_handler1)
        self.machine.events.add_handler('test_event_player2', self.event_handler2)
        self.machine.events.add_handler('test_event_player3', self.event_handler3)
        self.advance_time_and_run(1)

        self.machine.events.post('test_event_player1', test="123")
        self.advance_time_and_run(1)

        self.assertEqual(1, self._handler1_called)
        self.assertEqual(1, self._handler2_called)
        self.assertEqual(1, self._handler3_called)
        self.assertEqual(tuple(), self._handler1_args)
        self.assertEqual(dict(test="123"), self._handler1_kwargs)
        self.assertEqual(tuple(), self._handler2_args)
        self.assertEqual(dict(priority=0), self._handler2_kwargs)
        self.assertEqual(tuple(), self._handler3_args)
        self.assertEqual(dict(priority=0), self._handler3_kwargs)

    def test_event_player_delay(self):
        self.mock_event('test_event_player2')
        self.mock_event('test_event_player3')

        self.machine.events.post('test_event_player_delayed')
        self.machine_run()
        self.assertEqual(0, self._events['test_event_player2'])
        self.assertEqual(0, self._events['test_event_player3'])
        self.advance_time_and_run(2)
        self.assertEqual(1, self._events['test_event_player2'])
        self.assertEqual(1, self._events['test_event_player3'])

    def test_random_event_player(self):
        self.machine.events.add_handler('test_random_event_player1', self.event_handler1)
        self.machine.events.add_handler('test_random_event_player2', self.event_handler2)
        self.machine.events.add_handler('test_random_event_player3', self.event_handler3)
        self.advance_time_and_run(1)

        with patch('random.randint', return_value=1) as mock_random:
            self.machine.events.post('test_random_event_player1', test="123")
            self.advance_time_and_run(1)
            mock_random.assert_called_once_with(1, 2)

        self.assertEqual(1, self._handler1_called)
        self.assertEqual(1, self._handler2_called)
        self.assertEqual(0, self._handler3_called)
        self.assertEqual(tuple(), self._handler1_args)
        self.assertEqual(dict(test="123"), self._handler1_kwargs)
        self.assertEqual(tuple(), self._handler2_args)
        self.assertEqual(dict(test="123"), self._handler2_kwargs)

        with patch('random.randint', return_value=1) as mock_random:
            self.machine.events.post('test_random_event_player1', test="123")
            self.advance_time_and_run(1)
            mock_random.assert_called_once_with(1, 1)

        self.assertEqual(2, self._handler1_called)
        self.assertEqual(1, self._handler2_called)
        self.assertEqual(1, self._handler3_called)
        self.assertEqual(tuple(), self._handler1_args)
        self.assertEqual(dict(test="123"), self._handler1_kwargs)
        self.assertEqual(tuple(), self._handler3_args)
        self.assertEqual(dict(test="123"), self._handler3_kwargs)

    def test_event_player_in_mode(self):
        self.machine.events.add_handler('test_event_player_mode1', self.event_handler1)
        self.machine.events.add_handler('test_event_player_mode2', self.event_handler2)
        self.machine.events.add_handler('test_event_player_mode3', self.event_handler3)
        self.advance_time_and_run(1)

        # mode not loaded. event should not be replayed
        self.machine.events.post('test_event_player_mode1', test="123")
        self.advance_time_and_run(1)

        self.assertEqual(1, self._handler1_called)
        self.assertEqual(0, self._handler2_called)
        self.assertEqual(0, self._handler3_called)

        # start mode
        self.machine.events.post('test_mode_start')
        self.advance_time_and_run(1)
        self.assertTrue(self.machine.mode_controller.is_active("test_mode"))

        # now the event should get replayed
        self.machine.events.post('test_event_player_mode1', test="123")
        self.advance_time_and_run(1)

        self.assertEqual(2, self._handler1_called)
        self.assertEqual(1, self._handler2_called)
        self.assertEqual(1, self._handler3_called)

        # stop mode
        self.machine.events.post('test_mode_end')
        self.advance_time_and_run(1)
        self.assertFalse(self.machine.mode_controller.is_active("test_mode"))

        # event should not longer get replayed
        self.machine.events.post('test_event_player_mode1', test="123")
        self.advance_time_and_run(1)

        self.assertEqual(3, self._handler1_called)
        self.assertEqual(1, self._handler2_called)
        self.assertEqual(1, self._handler3_called)

    def test_random_event_player_in_mode(self):
        self.machine.events.add_handler('test_random_event_player_mode1', self.event_handler1)
        self.machine.events.add_handler('test_random_event_player_mode2', self.event_handler2)
        self.machine.events.add_handler('test_random_event_player_mode3', self.event_handler3)
        self.advance_time_and_run(1)

        # mode not loaded. event should not be replayed
        self.machine.events.post('test_random_event_player_mode1', test="123")
        self.advance_time_and_run(1)

        self.assertEqual(1, self._handler1_called)
        self.assertEqual(0, self._handler2_called)
        self.assertEqual(0, self._handler3_called)

        # start mode
        self.machine.events.post('test_mode_start')
        self.advance_time_and_run(1)
        self.assertTrue(self.machine.mode_controller.is_active("test_mode"))

        # now the event should get replayed
        with patch('random.randint', return_value=1) as mock_random:
            self.machine.events.post('test_random_event_player_mode1', test="123")
            self.advance_time_and_run(1)
            mock_random.assert_called_once_with(1, 2)

        self.assertEqual(2, self._handler1_called)
        self.assertEqual(1, self._handler2_called)
        self.assertEqual(0, self._handler3_called)

        # stop mode
        self.machine.events.post('test_mode_end')
        self.advance_time_and_run(1)
        self.assertFalse(self.machine.mode_controller.is_active("test_mode"))

        # event should not longer get replayed
        self.machine.events.post('test_random_event_player_mode1', test="123")
        self.advance_time_and_run(1)

        self.assertEqual(3, self._handler1_called)
        self.assertEqual(1, self._handler2_called)
        self.assertEqual(0, self._handler3_called)

    def event_block(self, queue, **kwargs):
        del kwargs
        self.queue = queue
        queue.wait()

    def test_random_event_player_in_game_mode(self):
        self.start_game()
        self.machine.events.add_handler('out1', self.event_handler2)
        self.machine.events.add_handler('out2', self.event_handler3)
        self.advance_time_and_run(1)

        # mode not loaded. event should not be replayed
        self.machine.events.post('test_random_event_player_mode2', test="123")
        self.advance_time_and_run(1)

        self.assertEqual(0, self._handler2_called)
        self.assertEqual(0, self._handler3_called)

        # start mode
        self.machine.events.post('game_mode_start')
        self.advance_time_and_run(1)
        self.assertTrue(self.machine.mode_controller.is_active("game_mode"))

        # now the event should get replayed
        with patch('random.randint', return_value=1) as mock_random:
            self.machine.events.post('test_random_event_player_mode2', test="123")
            self.advance_time_and_run(1)
            mock_random.assert_called_once_with(1, 2)

        self.assertEqual(1, self._handler2_called)
        self.assertEqual(0, self._handler3_called)

        # block stop of mode
        self.machine.events.add_handler('mode_game_mode_stopping', self.event_block)

        # stop game
        self.machine.game.stop()
        self.advance_time_and_run()

        # event should still work (and not crash)
        self.machine.events.post('test_random_event_player_mode2', test="123")
        self.machine_run()
        self.assertEqual(1, self._handler2_called)
        self.assertEqual(1, self._handler3_called)

        # when the block ends game should end too
        self.queue.clear()
        self.machine_run()

        # but no longer after clear
        self.machine.events.post('test_random_event_player_mode2', test="123")
        self.machine_run()

        self.advance_time_and_run(1)
        self.assertFalse(self.machine.mode_controller.is_active("game_mode"))

        self.assertEqual(1, self._handler2_called)
        self.assertEqual(1, self._handler3_called)

    def delay1_cb(self, **kwargs):
        del kwargs
        self.machine.events.post("event1")

    def event1_cb(self, **kwargs):
        del kwargs
        self.delay.add(ms=100, callback=self.delay2_cb)

    def delay2_cb(self, **kwargs):
        del kwargs
        self.machine.events.post("event2")

    def event2_cb(self, **kwargs):
        del kwargs
        self.delay.add(ms=100, callback=self.delay3_cb)

    def delay3_cb(self, **kwargs):
        del kwargs
        self.machine.events.post("event3")

    def event3_cb(self, **kwargs):
        del kwargs
        self.correct = True

    def test_event_in_delay(self):
        self.machine.events.add_handler('event1', self.event1_cb)
        self.machine.events.add_handler('event2', self.event2_cb)
        self.machine.events.add_handler('event3', self.event3_cb)
        self.correct = False
        self.delay = DelayManager(self.machine)

        self.machine.events.post("event1")
        self.advance_time_and_run(1)

        self.assertTrue(self.correct)

    def delay_first(self):
        self.called = True
        self.delay.remove("second")

    def delay_second(self):
        if not self.called:
            raise AssertionError("first has not been called")

        raise AssertionError("this should never be called")

    def test_delay_order(self):
        self.called = False
        self.delay = DelayManager(self.machine)

        self.delay.add(ms=6001, name="second", callback=self.delay_second)
        self.delay.add(ms=6000, name="first", callback=self.delay_first)

        self.advance_time_and_run(10)

    def delay_zero_ms(self, start):
        self.delay.add(ms=0, name="second", callback=self.delay_zero_ms_next_frame, start=start)

    def delay_zero_ms_next_frame(self, start):
        self.assertLessEqual(self.machine.clock.get_time(), start)

    def test_zero_ms_delay(self):
        self.called = False
        self.delay = DelayManager(self.machine)

        self.delay.add(ms=0, name="first", callback=self.delay_zero_ms, start=self.machine.clock.get_time())
        self.advance_time_and_run(10)

    def _handler(self, **kwargs):
        del kwargs
        self._called += 1

    def test_handler_with_condition(self):
        self._called = 0
        self.machine.events.add_handler("test{param > 1 and a == True}", self._handler)

        self.post_event("test")
        self.assertEqual(0, self._called)

        self.post_event_with_params("test", param=3, a=False)
        self.assertEqual(0, self._called)

        self.post_event_with_params("test", param=3, a=True)
        self.assertEqual(1, self._called)

    def test_handler_with_settings_condition_invalid_setting(self):
        self._called = 0
        self.machine.events.add_handler("test{settings.test == True}", self._handler)

        # invalid setting
        with self.assertRaises(AssertionError):
            self.post_event("test")
            self.assertEqual(0, self._called)

        # reset exception
        self._exception = None

    def test_handler_with_settings_condition(self):
        self._called = 0
        self.machine.events.add_handler("test{settings.test == True}", self._handler)


        self.machine.settings._settings = {}
        self.machine.settings.add_setting(SettingEntry("test", "Test", 1, "test", "a",
                                                       {False: "A (default)", True: "B"}))

        # setting false
        self.post_event("test")
        self.assertEqual(0, self._called)

        self.machine.settings.set_setting_value("test", True)

        # settings true
        self.post_event("test")
        self.assertEqual(1, self._called)

    def test_weighted(self):
        self.mock_event("out3")
        self.mock_event("out4")

        self.post_event("test_mode_start")
        self.advance_time_and_run()

        with patch('random.randint', return_value=1) as mock_random:
            self.machine.events.post('test_random_event_player_weighted')
            self.advance_time_and_run(1)
            mock_random.assert_called_once_with(1, 1001)

        self.assertEventCalled("out3")
        self.assertEventNotCalled("out4")

        self.mock_event("out3")
        self.mock_event("out4")

        with patch('random.randint', return_value=2) as mock_random:
            self.machine.events.post('test_random_event_player_weighted')
            self.advance_time_and_run(1)
            mock_random.assert_called_once_with(1, 1001)

        self.assertEventNotCalled("out3")
        self.assertEventCalled("out4")

        self.mock_event("out3")
        self.mock_event("out4")

        with patch('random.randint', return_value=500) as mock_random:
            self.machine.events.post('test_random_event_player_weighted')
            self.advance_time_and_run(1)
            mock_random.assert_called_once_with(1, 1001)

        self.assertEventNotCalled("out3")
        self.assertEventCalled("out4")

    def test_broken_handlers(self):
        # unbalanced placeholder
        with self.assertRaises(ValueError):
            self.machine.events.add_handler("event_name{machine.variables.test", self._handler)

        # unbalanced the other side
        with self.assertRaises(ValueError):
            self.machine.events.add_handler("event_namemachine.variables.test}", self._handler)

        # invalid space
        with self.assertRaises(ValueError):
            self.machine.events.add_handler("event_name {machine.variables.test}", self._handler)



"""Test case to start and stop games without ball devices."""
from unittest.mock import MagicMock

from mpf.tests.MpfGameTestCase import MpfGameTestCase


class MpfFakeGameTestCase(MpfGameTestCase):

    """Test case for a game that does not require ball devices & start switches.
    
    Often times you need to write a test that is able to start a game. However
    in order to start a game, MPF requires lots of things, like having proper
    ball devices and a start button and stuff like that.
    
    This test overwrites the ``start_game()`` and ``drain_ball()`` methods
    of the ``MpfGameTestCase`` class so that you can start games and drain
    balls without actually having any ball devices configured.
    
    """

    def __init__(self, methodName):
        """Patch minimal config into machine."""
        super().__init__(methodName)
        self.machine_config_patches['machine'] = dict()
        self.machine_config_patches['machine']['min_balls'] = 0

    def start_game(self, num_balls_known=3):
        """Start a game.
        
        Does not require ball devices or a start button to be present in the
        config file. Sets the number of known balls to 3.
        
        """
        def _add_ball(**kwargs):
            del kwargs
            self.machine.playfield.balls += 1
            self.machine.playfield.available_balls += 1

        # game start should work
        self.machine.playfield.add_ball = _add_ball
        self.machine.ball_controller.num_balls_known = num_balls_known
        super().start_game()

    def drain_all_balls(self):
        """Drain all the balls in play.
        
        Does not actually require any ball devices to be present in the config
        file.
        
        """
        drained_balls = 0
        for _ in range(self.machine.game.balls_in_play):
            result = self.post_relay_event_with_params("ball_drain", balls=1)
            drained_balls += result['balls']
        self.machine.playfield.balls -= drained_balls
        self.machine.playfield.available_balls -= drained_balls
        self.advance_time_and_run()

    def drain_one_ball(self):
        """Drain one ball.

        Does not actually require any ball devices to be present in the config
        file.

        """
        result = self.post_relay_event_with_params("ball_drain", balls=1)
        self.machine.playfield.balls -= result['balls']
        self.machine.playfield.available_balls -= result['balls']
        self.advance_time_and_run()


"""Python logging formatters."""
import json
import logging
import base64
from datetime import datetime


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code."""
    if isinstance(obj, datetime):
        serial = obj.isoformat()
        return serial
    raise TypeError("Type not serializable")


def format_time(record, datefmt=None):
    """Override default to use strftime, e.g. to get microseconds."""
    created = datetime.fromtimestamp(record.created)
    if not datefmt:
        datefmt = "%Y-%m-%dT%H:%M:%S.{:03f}%z".format(record.msecs)
    return created.strftime(datefmt)


class JSONFormatter(logging.Formatter):

    """A formatter that renders log records as JSON objects.

    Format: {"timestamp":"...", "level":"...", "name":"...", "message":"..."}
    """

    def format(self, record):
        """Encode log record as JSON."""
        log = {
            'timestamp': format_time(record, self.datefmt),
            'level': record.levelname,
            'name': record.name,
            'message': record.getMessage(),
            'ident': 'ServiceLog',
        }
        if record.exc_info:
            try:
                traceback = self.formatException(record.exc_info)
                traceback = base64.b64encode(traceback.encode('utf-8'))
                traceback = traceback.decode('utf-8')
            except TypeError:
                traceback = 'unable to serialize exception'
            log['traceback'] = traceback
        if isinstance(record.msg, dict):
            log['json_message'] = record.msg

        log = json.dumps(log, default=json_serial)
        return log



"""Command to start the MPF game engine."""

import argparse
import errno
import os
import signal
import socket
import sys
from datetime import datetime
import logging
from logging.handlers import QueueHandler, SysLogHandler

from queue import Queue

from mpf.core.machine import MachineController
from mpf.core.utility_functions import Util
from mpf.core.config_loader import YamlMultifileConfigLoader, ProductionConfigLoader
from mpf.commands.logging_formatters import JSONFormatter


class Command:

    """Runs the mpf game."""

    # pylint: disable-msg=too-many-locals,too-many-statements
    def __init__(self, mpf_path, machine_path, args):
        """Run mpf game."""
        del mpf_path
        self.machine = None
        self._sigint_count = 0

        parser = argparse.ArgumentParser(
            description='Starts the MPF game engine')

        parser.add_argument("-a",
                            action="store_true", dest="no_load_cache",
                            help="Forces the config to be loaded from files "
                                 "and not cache")

        parser.add_argument("-A",
                            action="store_false", dest="create_config_cache",
                            help="Does not create the cache config files")

        parser.add_argument("-b",
                            action="store_false", dest="bcp", default=True,
                            help="Runs MPF without making a connection "
                                 "attempt to a "
                                 "BCP Server")

        parser.add_argument("-c",
                            action="store", dest="configfile",
                            default="config.yaml", metavar='config_file',
                            help="The name of a config file to load. Default "
                                 "is "
                                 "config.yaml. Multiple files can be used "
                                 "via a comma-"
                                 "separated list (no spaces between)")

        parser.add_argument("-f",
                            action="store_true", dest="force_assets_load",
                            default=False,
                            help="Load all assets upon startup.  Useful for "
                            "ensuring all assets are set up properly "
                            "during development.")

        parser.add_argument("--json-logging",
                            action="store_true", dest="jsonlogging",
                            default=False,
                            help="Enables json logging to file. ")

        parser.add_argument("-l",
                            action="store", dest="logfile",
                            metavar='file_name',
                            default=os.path.join(
                                "logs",
                                datetime.now().strftime(
                                    "%Y-%m-%d-%H-%M-%S-mpf-" +
                                    socket.gethostname() + ".log")),
                            help="The name (and path) of the log file")

        parser.add_argument("-p",
                            action="store_true", dest="pause", default=False,
                            help="Pause the terminal window on exit. Useful "
                            "when launching in a separate window so you can "
                            "see any errors before the window closes.")

        parser.add_argument("-P",
                            action="store_true", dest="production", default=False,
                            help="Production mode. Will suppress errors, wait for hardware on start and "
                                 "try to exit when startup fails. Run this inside a loop.")

        parser.add_argument("-t",
                            action="store_false", dest='text_ui', default=True,
                            help="Use the ASCII test-based UI")

        parser.add_argument("-v",
                            action="store_const", dest="loglevel",
                            const=logging.DEBUG,
                            default=15,
                            help="Enables verbose logging to the"
                                 " log file")

        parser.add_argument("-V",
                            action="store_const", dest="consoleloglevel",
                            const=logging.DEBUG,
                            default=logging.INFO,
                            help="Enables verbose logging to the console. DO "
                                 "NOTE: On Windows platforms you must also use -v for "
                                 "this to work.")

        parser.add_argument("-x",
                            action="store_const", dest="force_platform",
                            const='virtual',
                            help="Forces the virtual platform to be "
                                 "used for all devices")

        parser.add_argument("--vpx",
                            action="store_const", dest="force_platform",
                            const='virtual_pinball',
                            help="Forces the virtual_pinball platform to be "
                                 "used for all devices")

        parser.add_argument("--syslog_address",
                            action="store", dest="syslog_address",
                            help="Log to the specified syslog address. This "
                                 "can be a domain socket such as /dev/og on "
                                 "Linux or /var/run/syslog on Mac. "
                                 "Alternatively, you an specify host:port for "
                                 "remote logging over UDP.")

        parser.add_argument("-X",
                            action="store_const", dest="force_platform",
                            const='smart_virtual',
                            help="Forces the smart virtual platform to be "
                                 "used for all"
                                 " devices")

        # The following are just included for full compatibility with mc
        # which is needed when using "mpf both".

        parser.add_argument("-L",
                            action="store", dest="mc_file_name",
                            metavar='mc_file_name',
                            default=None, help=argparse.SUPPRESS)

        parser.add_argument("--no-sound",
                            action="store_true", dest="no_sound", default=False)

        self.args = parser.parse_args(args)
        self.args.configfile = Util.string_to_event_list(self.args.configfile)

        # Configure logging. Creates a logfile and logs to the console.
        # Formatting options are documented here:
        # https://docs.python.org/2.7/library/logging.html#logrecord-attributes

        try:
            os.makedirs(os.path.join(machine_path, 'logs'))
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise

        full_logfile_path = os.path.join(machine_path, self.args.logfile)

        try:
            os.remove(full_logfile_path)
        except OSError:
            pass

        if self.args.text_ui:
            console_log = logging.NullHandler()
            console_log.setLevel(logging.ERROR)
        else:
            console_log = logging.StreamHandler()
            console_log.setLevel(self.args.consoleloglevel)

        # tell the handler to use this format
        console_log.setFormatter(logging.Formatter(
            '%(levelname)s : %(name)s : %(message)s'))

        # initialise async handler for console
        console_log_queue = Queue()
        console_queue_handler = QueueHandler(console_log_queue)
        self.console_queue_listener = logging.handlers.QueueListener(
            console_log_queue, console_log)
        self.console_queue_listener.start()

        # initialise file log
        file_log = logging.FileHandler(full_logfile_path)
        if self.args.jsonlogging:
            formatter = JSONFormatter()
        else:
            formatter = logging.Formatter('%(asctime)s : %(levelname)s : %(name)s : %(message)s')
        file_log.setFormatter(formatter)

        # initialise async handler for file log
        file_log_queue = Queue()
        file_queue_handler = QueueHandler(file_log_queue)
        self.file_queue_listener = logging.handlers.QueueListener(
            file_log_queue, file_log)
        self.file_queue_listener.start()

        # add loggers
        logger = logging.getLogger()
        logger.addHandler(console_queue_handler)
        logger.addHandler(file_queue_handler)
        logger.setLevel(self.args.loglevel)
        logger.info("Loading config.")

        if self.args.syslog_address:
            try:
                host, port = self.args.syslog_address.split(":")
            except ValueError:
                syslog_logger = SysLogHandler(self.args.syslog_address)
            else:
                syslog_logger = SysLogHandler((host, int(port)))

            logger.addHandler(syslog_logger)

        signal.signal(signal.SIGINT, self.sigint_handler)

        if not self.args.production:
            config_loader = YamlMultifileConfigLoader(machine_path, self.args.configfile,
                                                      not self.args.no_load_cache, self.args.create_config_cache)
        else:
            config_loader = ProductionConfigLoader(machine_path)

        config = config_loader.load_mpf_config()

        try:
            self.machine = MachineController(vars(self.args), config)
            self.machine.add_crash_handler(self.restore_logger)
            self.machine.run()
            logging.info("MPF run loop ended.")
            self.exit()

        # pylint: disable-msg=broad-except
        except Exception as e:
            self.exit(exception=e)

    def sigint_handler(self, signum=None, frame=None):
        """Handle SIGINT."""
        del signum, frame
        self._sigint_count += 1
        if self._sigint_count > 1:
            self.exit("Received second SIGINT. Will exit ungracefully!")
        elif self.machine:
            self.machine.stop("SIGINT or keyboard interrupt")
        else:
            self.exit("Shutdown because of SIGINT or keyboard interrupt.")

    def restore_logger(self):
        """Restore logger."""
        if self.args.text_ui:
            # Re-enable console logging
            logger = logging.getLogger()
            logger.addHandler(logging.StreamHandler())

    def exit(self, exception=None):
        """Handle MPF exit from either a clean shutdown or from a crash.

        Cleanly shuts down logging and restores the console window if the Text
        UI option is used.
        """
        if exception:
            logging.exception(exception)

        logging.shutdown()
        self.console_queue_listener.stop()
        self.file_queue_listener.stop()

        if self.args.pause:
            input('Press ENTER to continue...')     # nosec

        sys.exit()
